from __future__ import annotations

from model.Reaction import Reaction
from model.MoleculeType import MoleculeType
import math
import decimal

    
    
def add_all_mentioned_products(molecules:set[MoleculeType], reactions:list[Reaction]) -> set[MoleculeType]:
    res = molecules
    for reaction in reactions:
        if reaction.direction in {"forward", "both"}:
            res = res.union(reaction.products)
        if reaction.direction in {"reverse", "both"}:
            res = res.union(reaction.reactants)
    return res

def compute_closure(molecules:set[MoleculeType], reactions:list[Reaction]) -> set[MoleculeType]:
    all_molecules = molecules.copy()
    size = -1
    while len(all_molecules) > size:
        size = len(all_molecules)
        for reaction in reactions:
            if reaction.direction in {"forward", "both"}:
                if reaction.reactants.issubset(all_molecules):
                    all_molecules.update(reaction.products)
            if reaction.direction in {"reverse", "both"}:
                if reaction.products.issubset(all_molecules):
                    all_molecules.update(reaction.reactants)
    
    return all_molecules

def filter_reactions(food:set[MoleculeType], reactions:list[Reaction]) -> list[Reaction]:
    res_reactions = []
    for r in reactions:
        if r.is_catalyzed_uninhibited_all_reactants(r.direction, food=food):
            res_reactions.append(r)
    return res_reactions

def compute_food_generated(food:set[MoleculeType], reactions:list[Reaction]) -> list[Reaction]:
    available_food = food
    available_reactions = reactions
    closure = []
    while(True):
        to_add = [r for r in available_reactions if r.is_all_reactants(food=available_food, direction=r.direction)]
        if len(to_add) > 0:
            closure.extend(to_add)
            for reaction in to_add:
                match reaction.direction:
                    case "forward":
                        available_food.update(reaction.products)
                    case "reverse":
                        available_food.update(reaction.reactants)
                    case "both":
                        res = set()
                        if reaction.reactants.issubset(available_food):
                            res.update(reaction.products)
                        if reaction.products.issubset(available_food):
                            res.update(reaction.reactants)
                        available_food.update(res)
            for reaction in to_add: available_reactions.remove(reaction)
        else:
            break
    return closure

def is_float(s:str)-> bool:
    return s.replace(".", "", 1).replace(",", "", 1).isdigit()

def parse_float_input(input: str) -> list | range:

    if "," in input:
        output_ints = input.split(",")
        output_ints = [o.strip() for o in output_ints]
        output = []
        for o in output_ints:
            if is_float(o) and not o.startswith("-", 0, 1):
                output.append(float(o))
    elif "-" in input:
        output = []
        if "/" in input:
            output_ints = [input.split(
                "-")[0]].extend(input.split("-")[1].split("/"))
            while output_ints[0] <= output_ints[1]:
                output.append(output_ints[0])
                output_ints[0] += output_ints[2]
        else:
            output_ints = input.split("-")
            while output_ints[0] <= output_ints[1]:
                output.append(output_ints[0])
                output_ints[0] += 1.0
    else:
        output = [(float(input) if not input.strip().startswith("-", 0, 1)
                and is_float(input.strip()) else None)]

    return output

def parse_integer_input(input: str) -> list | range:

    if "," in input:
        output_ints = input.split(",")
        output = []
        for o in output_ints:
            if o.strip().isdigit():
                output.append(int(o.strip()))
    elif "-" in input:
        if "/" in input:
            output_ints = [input.split("-")[0]]
            output_ints.extend(input.split("-")[1].split("/"))
            output_ints = [int(output_int) for output_int in output_ints]
            output = range(output_ints[0], output_ints[1], output_ints[2])
        else:
            output_ints = input.split("-")
            output_ints = [int(output_int) for output_int in output_ints]
            output = range(output_ints[0], output_ints[1])
    else:
        output = [int(input) if input.strip().isdigit() else None]

    return output

def binom_dist(n: int, p: float) -> list[float]:

    dist = []
    p_of_eq_k = []
    for k in range(n+1):
        p_of_eq_k.append(decimal.Decimal(math.comb(n, k)) *
                         decimal.Decimal(p**k) * decimal.Decimal((1-p)**(n-k)))
        dist.append(sum(p_of_eq_k))
        if sum(p_of_eq_k) == sum(p_of_eq_k[:-1]):
            break 
    return dist

def replace_parameters(input_str: str, 
                       parameters: dict[str:int, str:int, str:int, str:float, str:int]) -> str:
    
    if "." in str(parameters['m']):
        mean_str = str(parameters['m']).replace('.', "-")
    else:
        mean_str = str(parameters['m'])
        
    return (input_str.replace("#a", str(parameters["a"]))
            .replace("#k", str(parameters["k"]))
            .replace("#n", str(parameters["n"]))
            .replace("#m", mean_str)
            .replace("#r", str(parameters["r"])))


from __future__ import annotations
from algorithm.AlgorithmBase import AlgorithmBase
from model.ReactionSystem import ReactionSystem
import Utilities as Utilities

import sys
import os
sys.path.insert(0, os.path.abspath(
    os.path.join(os.path.dirname(__file__), '.')))


class MaxCAFAlgorithm(AlgorithmBase):

    NAME: str = "Max CAF"

    @property
    def name(self):
        return self.NAME

    @name.setter
    def name(self, value: str):
        self.NAME = value

    @property
    def description(self):
        return "computes the maximal CAF [HMS15]"

    def apply(self, input: ReactionSystem) -> ReactionSystem:
        """Computes the maximal CAF

        Args:
            input (ReactionSystem): system to be computed

        Returns:
            ReactionSystem: maximal CAF
        """        
        result = ReactionSystem(self.NAME)

        input_reactions = set(input.reactions)
        input_food = input.foods

        molecules = [input_food]
        reactions = [Utilities.filter_reactions(input_food, input_reactions)]

        i = 0
        molecules.insert(
            i+1, Utilities.add_all_mentioned_products(molecules[i], 
                                                      reactions[i]))
        reactions.insert(
            i+1, Utilities.filter_reactions(molecules[i+1], input_reactions))

        while len(reactions[i+1]) > len(reactions[i]):
            i += 1
            molecules.insert(
                i+1, Utilities.add_all_mentioned_products(molecules[i], 
                                                          reactions[i]))
            reactions.insert(
                i+1, Utilities.filter_reactions(molecules[i+1], 
                                                input_reactions))

        if len(reactions[i+1]) > 0:
            result.reactions = reactions[i+1]
            result.foods = result.compute_mentioned_foods(input.foods)
        return result



def compute(expression: str) -> str:
    """
    Return a Disjunctive Normal Form of the expression as a str.

    The resulting str uses "," as "or"/disjunction and "&" as 
    "and"/conjunction and does not contain any brackets.
    The input-expression can contain the logical operators:
        "or" as ","
        "and" as "&"
        brackets as "(" and ")" 
    Everything else will be treated as elements.
    There cannot be any whitespaces in the input-expression.

    Parameters:
        expression  (str): A logical expression with brackets to be turned
        into a DNF without brackets

    Returns:

    """

    return ",".join(recurse(expression))


def recurse(expression: str) -> list[str]:
    last_pos = len(expression)
    if expression.startswith("("):
        associated_closed_bracket = find_associated_closed_bracket(
            expression, 0)
        if associated_closed_bracket+1 == last_pos:
            return recurse(expression[1:associated_closed_bracket])
        else:
            next_outside_operator = expression[associated_closed_bracket + 1]
            if next_outside_operator == ",":
                return union(recurse(expression[1:associated_closed_bracket]),
                             recurse(expression[associated_closed_bracket + 2:last_pos]))
            if next_outside_operator == "&":
                next_rel_or = next_outside_or(
                    expression, associated_closed_bracket+1)
                if next_rel_or == -1:
                    return product(recurse(expression[1:associated_closed_bracket]),
                                   recurse(expression[associated_closed_bracket + 2:last_pos]))
                else:
                    return union(product(recurse(expression[1:associated_closed_bracket]),
                                         recurse(expression[associated_closed_bracket + 2:next_rel_or])),
                                 recurse(expression[next_rel_or+1:]))
    else:
        next_inside_operator = next_or(expression, 0)
        if next_inside_operator > 0:
            return union(recurse(expression[0:next_inside_operator]),
                         recurse(expression[next_inside_operator + 1:last_pos]))
        next_inside_operator = next_and(expression, 0)
        if next_inside_operator > 0:
            second_expression = expression[next_inside_operator + 1:last_pos]
            end_of_brackets = find_associated_closed_bracket(
                expression, next_inside_operator+1)
            if end_of_brackets > 0:
                position = end_of_brackets + 1
                while position < last_pos:
                    if expression[position] == ",":
                        return union(product(recurse(expression[0:next_inside_operator]),
                                             recurse(expression[next_inside_operator + 1:position])),
                                     recurse(expression[position + 1:last_pos]))
                    elif expression[position] == "(":
                        position = find_associated_closed_bracket(
                            expression, position)
                    position += 1
            return product(recurse(expression[0:next_inside_operator]),
                           recurse(expression[next_inside_operator + 1:last_pos]))

    return [expression]


def union(tree_a: list, tree_b: list) -> list:
    tree_a.extend(tree_b)
    return tree_a


def product(tree_a: list, tree_b: list) -> list:
    res = []
    for content_a in tree_a:
        for content_b in tree_b:
            if content_a == content_b:
                res.append(content_a)
            else:
                res.append(content_a + "&" + content_b)
    return res


def next_outside_or(expression: str, start_pos: int) -> int:
    """Finds the next or on the bracket-level of startpos.

    Args:
        expression (str): expression to be searched through
        start_pos (int): position at start of search in expression

    Returns:
        start_pos (int): position of next ','
        -1 (int): if ',' is not found at bracket-level
    """
    end_pos = len(expression)-1
    inside_brackets_depth = 0
    while start_pos <= end_pos:
        if inside_brackets_depth == 0:
            if start_pos == end_pos:
                return -1
            elif expression[start_pos] == ",":
                return start_pos
        if expression[start_pos] == "(":
            inside_brackets_depth += 1
        elif expression[start_pos] == ")":
            inside_brackets_depth -= 1
        start_pos += 1

    return -1


def next_or(expression: str, start_pos: int) -> int:
    """
    Return the postion of the next "," after "start_pos" in "expression".

    If no "," is found returns -1 instead.
    If no "," is found before the next open bracket "(" returns -1 as well.

    Parameters:
        expression  (str): The string to be searched through
        start_pos   (int): The position at which searching starts.

    Returns:
        start_pos   (int): Position of the next ","
        -1          (int): Returned if no relevant "," is found
    """
    end_pos = len(expression) - 1
    while start_pos <= end_pos:
        if expression[start_pos] == ",":
            return start_pos
        elif expression[start_pos] == "(":
            return -1
        start_pos += 1

    return -1


def next_and(expression: str, start_pos: int) -> int:
    """
    Return the postion of the next "&" after "start_pos" in "expression".

    If no "&" is found returns -1 instead.
    If no "&" is found before the next open bracket "(" returns -1 as well.

    Parameters:
        expression  (str): The string to be searched
        start_pos   (int): The position at which searching starts.

    Returns:
        start_pos   (int): Position of the next "&"
        -1          (int): Returned if no relevant "&" is found
    """
    end_pos = len(expression) - 1
    while start_pos <= end_pos:
        if expression[start_pos] == "&":
            return start_pos
        elif expression[start_pos] == "(":
            return -1
        start_pos += 1

    return -1


def find_associated_closed_bracket(expression: str, start_pos: int) -> int:
    """
    Return position of associated bracket of the open bracket at "start_pos".

    If "start_pos" is not the position of an open bracket returns start_pos.
    If no closed bracket is found after an open bracket returns -1 instead.

    Parameters:
        expression  (str): The string to be analyzed
        start_pos   (int): The position at which analysis starts.

    Returns:
        start_pos   (int): Either the initial position or the position of
        the associated closed bracket.
        -1          (int): Returned if the expression doesn't contain an
        associated bracket
    """
    end_pos = len(expression) - 1
    depth = 0
    while start_pos <= end_pos:
        if expression[start_pos] == "(":
            depth += 1
        elif expression[start_pos] == ")":
            depth -= 1
        if depth == 0:
            return start_pos
        start_pos += 1
    return -1



from itertools import combinations_with_replacement
import math
import catPyNet.main.CatPyNet as cpn
def main():
    
    alph = ['a', 'b']
    for i in range(11):
        print("Length of a=" + str(len(alph)) + "k=" + str(i)+ " " + str(len(list(combinations_with_replacement(alph, i)))))
        print(str(sum([j for j in  range(len(alph) + i)])))

def get_org(pathway_code:str, org_code:str, output_path:str):
    import requests
    import time
    pathway_list = []
    with open("G:\\Github\\BA-Jan\\test_files\\eco00010             Glycolysis  Gl.txt", 'r') as f:
        lines = f.readlines()
        for line in lines:
            pathway_list.append(line.split()[0].strip())
    
    reactions = []
    all_modules = []
    for pathway_code in pathway_list:
        url = f'https://rest.kegg.jp/get/{pathway_code}'
        response = requests.get(url)
        time.sleep(0.334)    
        if response.status_code == 200:
            pathways_str = response.text
            modules = []
            for pthw_line in pathways_str.splitlines():
                if pthw_line.startswith('MODULE'):
                    modules.append(pthw_line.removeprefix('MODULE').split()[0].strip())
                elif pthw_line.strip().startswith(org_code + '_M'):
                    modules.append(pthw_line.split()[0].strip())
            reaction_strs = set()
            for mod in modules:
                mod = mod.removeprefix(org_code + '_')
                if mod in all_modules: continue
                all_modules.append(mod)
                print(mod)
                url = f'https://rest.kegg.jp/get/{mod}'
                response = requests.get(url)
                time.sleep(0.334)
                mod_lines = response.text.splitlines()
                reaction_bool = False
                for mod_line in mod_lines:
                    if mod_line.startswith('REACTION'):
                        reaction_bool = True
                    elif mod_line.startswith('COMPOUND'):
                        break
                    if reaction_bool:
                        mod_line = mod_line.removeprefix('REACTION').strip()
                        try: reaction_str = mod_line.split()[0].strip()
                        except: continue
                        sub_reaction_strs = []
                        if ',' in reaction_str:
                            sub_reaction_strs.extend(reaction_str.split(','))
                        if not sub_reaction_strs:
                            reaction_strs.update([reaction_str])
            for reaction_str in reaction_strs:
                if reaction_str in [react[0] for react in reactions]: continue
                url = f'https://rest.kegg.jp/get/{reaction_str}'
                response = requests.get(url)
                time.sleep(0.334)
                reac_lines = response.text.splitlines()
                reaction = [reaction_str, "",[]]
                for reac_line in reac_lines:
                    if reac_line.startswith('EQUATION'):
                        reaction[1] = reac_line.removeprefix('EQUATION').strip()
                    elif reac_line.startswith('ENZYME'):
                        reaction[2] = [enz.strip() for enz in reac_line.split()[1:]]
                        break
                print(str(reaction))
                reactions.append(reaction)
                """ for enzymes in [reaction[1] for reaction in reactions]:
                    for enzyme in enzymes:
                        enz_url = f'https://rest.kegg.jp/get/{enzyme}'
                        response = requests.get(enz_url)
                        time.sleep(0.334)
                        print(response.text) """     
        else:
            print(f"Failed to retrieve data. Status code: {response.status_code}")
    if reactions:
        reaction_obj_lst:list[Reaction] = []
        from catPyNet.model.ReactionSystem import Reaction, ReactionSystem
        for i, reaction in enumerate(reactions):
            reaction_obj_lst.insert(i, Reaction().parse_new(" : ".join([reaction[0], reaction[1]]),False))
            reaction_obj_lst[i].catalysts = ",".join(reaction[2])
        output_rs = ReactionSystem(reactions=reaction_obj_lst)
        cpn.export_rs_to_output_file([output_rs], output_path)
        total_res_rs = cpn.parse_input_file_to_rs("G:\\Github\\BA-Jan\\test_results\\test_reactions.crs")
        shared_reactions = set(total_res_rs.reactions).intersection(output_rs.reactions)
        output_rs = shared_reactions
        cpn.export_rs_to_output_file([output_rs], "G:\\Github\\BA-Jan\\test_results\e_coli_rs_shared.crs")
    else: print('no reactions')
            

def compare_files():
    sys_1 = cpn.parse_input_file_to_rs("G:\\Github\\BA-Jan\\test_results\\test_reactions.crs")
    sys_2 = cpn.parse_input_file_to_rs("G:\\Github\\BA-Jan\\test_results\\e_coli_rs.crs")
    shared_reactions = set([reaction.name for reaction in sys_2.reactions]).intersection([reaction.name for reaction in sys_1.reactions])
    print([reaction.name for reaction in sys_2.reactions])
    print([reaction.name for reaction in sys_1.reactions])
    print([reaction for reaction in shared_reactions])
    #sys_1.reactions = shared_reactions
    cpn.export_rs_to_output_file(sys_1, "G:\\Github\\BA-Jan\\test_results\e_coli_rs_shared.crs")
    
if __name__ == "__main__":
    """ import requests
    enzyme = '5.3.1.1'
    enz_url = f'https://rest.kegg.jp/get/{enzyme}'
    response = requests.get(enz_url)
    print(response.text) """
    #get_org(pathway_code= 'eco01100', org_code='eco', output_path="G:\\Github\\BA-Jan\\test_results\e_coli_rs.crs")
    compare_files()


# -*- coding: utf-8 -*-

import sys
import os
import time
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
application_path = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))


from hypothesis import assume, example, given, settings, strategies as st
import unittest
import copy

from fileIO.ModelIO import ModelIO
from settings.ReactionNotation import ReactionNotation
from settings.ArrowNotation import ArrowNotation
from model.MoleculeType import MoleculeType
from model.ReactionSystem import ReactionSystem
from model.Reaction import Reaction
from model.DisjunctiveNormalForm import *


listsize = 3
base_alphabet = ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z']

def standard_list(strat):
    return st.lists(strat, min_size=listsize, max_size=listsize)

def standard_text():
    return st.text(base_alphabet, min_size=1)

class BasicTestSuite(unittest.TestCase):
    """Basic test cases."""

    def test_absolute_truth_and_meaning(self):
        assert True
        
class MoleculeTypeTests(unittest.TestCase):
    """Tests for Moleculetype"""
    
    def test_basic_generation(self):
        
        result = MoleculeType()
        self.assertEqual(MoleculeType(), result)
        
    def test_value_generation(self):
        
        data = "A"
        result = MoleculeType(data)
        self.assertEqual(MoleculeType(data), result)
        
    def test_value_generation_without_constructor(self):
        
        data = MoleculeType()
        result_of_function = data.value_of("A")
        resulting_obj = data
        self.assertEqual(MoleculeType(name2type={"A":MoleculeType("A")}), resulting_obj)
        self.assertEqual(MoleculeType("A"), result_of_function)

    def test_values_generation_without_constructor(self):
        
        data_obj = MoleculeType("A", {"A":MoleculeType("A")})
        data_list = ["A", "B", "C", "A"]
        result_of_function = data_obj.values_of(data_list)
        self.assertEqual(set([MoleculeType("A"), MoleculeType("B"), MoleculeType("C"), MoleculeType("A")]), result_of_function)
        self.assertEqual(MoleculeType("A", {"A":MoleculeType("A"), "B":MoleculeType("B"), "C":MoleculeType("C")}), data_obj)
        
    def test_copy(self):
        
        data = MoleculeType("A", {"A":MoleculeType("A")})
        result = copy.deepcopy(data)
        self.assertEqual(data, result)

class ReactionTests(unittest.TestCase):
    """Tests for Reaction"""
    
    def __init__(self, methodName: str = "runTest") -> None:
        self.BASIC_REACTION = Reaction("Test_Reaction", warned_about_suppressing_coefficients = True,
                                reactants = MoleculeType().values_of(["A", "B"]), products = MoleculeType().values_of(["Z", "Y"]),
                                catalysts = "C,D", inhibitions =  MoleculeType().values_of(["E"]),
                                reactant_coefficients = {"A":1, "B":2},
                                product_coefficients = {"Z":1, "Y": 1},
                                direction = "forward")
        self.EMPTY_REACTION = Reaction()
    
        super().__init__(methodName)
    
    def basic_to_reaction(self,name:str, warned:bool, reactants:list[str], products:list[str], catalysts:list[str]
                              , inhibitors:list[str], reaction_coefficients:list[int], product_coefficients:list[int],
                              direction)->Reaction:
        reactants_m = MoleculeType().values_of(reactants)
        products_m = MoleculeType().values_of(products)
        inhibitors_m = MoleculeType().values_of(inhibitors)
        catalysts_str = catalysts[0]
        for i, c in enumerate(catalysts):
            if i == 0:
                continue
            catalysts_str += "," + c
        reaction_coefficients_dict = {mol:reaction_coefficients[i] for i, mol in enumerate(reactants_m)}
        product_coefficients_dict = {mol:product_coefficients[i] for i, mol in enumerate(products_m)}
        match str(direction):
            case "0":
                direction_key = "forward"
            case "1":
                direction_key = "reverse"
            case "2":
                direction_key = "both"
        test_reaction = Reaction(name, warned_about_suppressing_coefficients=warned, reactants = reactants_m,
                                       products=products_m, catalysts = catalysts_str, inhibitions=inhibitors_m,
                                       reactant_coefficients=reaction_coefficients_dict,
                                       product_coefficients=product_coefficients_dict, direction=direction_key)
        
        return test_reaction
    
    @given(name=st.text(base_alphabet, min_size=1), warned = st.booleans(), reactants = standard_list(standard_text()), 
           products = standard_list(standard_text()), catalysts = standard_list(standard_text()), inhibitors = standard_list(standard_text()),
           reactant_coefficients = st.lists(st.integers(min_value=1), min_size=listsize, max_size=listsize), product_coefficients = st.lists(st.integers(min_value=1), min_size=listsize, max_size=listsize),
           direction = st.integers(0, 2))
    def test_basic_generation(self, name:str, warned:bool, reactants:list[str], products:list[str], catalysts:list[str]
                              , inhibitors:list[str], reactant_coefficients:list[int], product_coefficients:list[int],
                              direction):
        
        result = self.basic_to_reaction(name, warned, reactants, products, catalysts, inhibitors, reactant_coefficients, product_coefficients, direction)
        same_result = self.basic_to_reaction(name, warned, reactants, products, catalysts, inhibitors, reactant_coefficients, product_coefficients, direction)
        self.assertEqual(same_result, result)
        
    def test_name_generation(self):
        
        result = Reaction("A")
        expected = Reaction("A", warned_about_suppressing_coefficients = False,
                                reactants = set(), products = set(),
                                catalysts = "", inhibitions = set(),
                                reactant_coefficients = {},
                                product_coefficients = {},
                                direction = "forward")
        self.assertEqual(result, expected)
        
    @given(name=st.text(base_alphabet, min_size=1), warned = st.booleans(), reactants = standard_list(standard_text()), 
           products = standard_list(standard_text()), catalysts = standard_list(standard_text()), inhibitors = standard_list(standard_text()),
           reactant_coefficients = st.lists(st.integers(min_value=1), min_size=listsize, max_size=listsize), product_coefficients = st.lists(st.integers(min_value=1), min_size=listsize, max_size=listsize),
           direction = st.integers(0, 2))
    def test_copy(self, name:str, warned:bool, reactants:list[str], products:list[str], catalysts:list[str]
                              , inhibitors:list[str], reactant_coefficients:list[int], product_coefficients:list[int],
                              direction):
        
        expected = self.basic_to_reaction(name, warned, reactants, products, catalysts, inhibitors, reactant_coefficients, product_coefficients, direction)
        result = copy.deepcopy(expected)
        self.assertEqual(expected, result)
    
    @given(name=st.text(base_alphabet, min_size=1), warned = st.booleans(), reactants = standard_list(standard_text()), 
           products = standard_list(standard_text()), catalysts = standard_list(standard_text()), inhibitors = standard_list(standard_text()),
           reactant_coefficients = st.lists(st.integers(min_value=1), min_size=listsize, max_size=listsize), product_coefficients = st.lists(st.integers(min_value=1), min_size=listsize, max_size=listsize),
           direction_reaction = st.integers(0, 2), direction_test = st.integers(0, 2), food = st.lists(st.text(base_alphabet, min_size=1)))
    def test_inhibitions_catalization_food_catalyzed(self, name:str, warned:bool, reactants:list[str], products:list[str], catalysts:list[str]
                              , inhibitors:list[str], reactant_coefficients:list[int], product_coefficients:list[int],
                              direction_reaction, direction_test, food:list[str]):
        
        data_obj:Reaction = self.basic_to_reaction(name, warned, reactants, products, catalysts, inhibitors, reactant_coefficients, product_coefficients, direction_reaction)
        data_food_catalyzed = MoleculeType().values_of(food)
        data_catalysts = MoleculeType().values_of(data_obj.catalysts.split(","))
        match str(direction_test):
            case "0":
                direction_key = "forward"
            case "1":
                direction_key = "reverse"
            case "2":
                direction_key = "both"
        result_catalyzed = data_obj.is_catalyzed_uninhibited_all_reactants(direction_key, food=data_food_catalyzed)
        
        if (direction_key in {"forward" or "both"}
            and data_food_catalyzed 
            and data_obj.reactants.issubset(data_food_catalyzed)
            and data_catalysts.issubset(data_food_catalyzed)
            and data_obj.inhibitions.isdisjoint(data_food_catalyzed)):
            self.assertTrue(result_catalyzed)
        elif (direction_key in {"forward" or "both"}
              and data_food_catalyzed 
            and data_obj.products.issubset(data_food_catalyzed)
            and data_catalysts.issubset(data_food_catalyzed)
            and data_obj.inhibitions.isdisjoint(data_food_catalyzed)):
            self.assertTrue(result_catalyzed)
        elif (direction_key in {"forward" or "both"}
              and data_food_catalyzed 
            and data_obj.reactants.issubset(data_food_catalyzed) 
            and data_catalysts.issubset(data_food_catalyzed)
            and data_obj.inhibitions.issubset(data_food_catalyzed)
            and data_obj.inhibitions):
            self.assertFalse(result_catalyzed)
        elif (direction_key in {"forward" or "both"}
              and data_food_catalyzed 
            and data_obj.products.issubset(data_food_catalyzed)
            and data_catalysts.issubset(data_food_catalyzed)
            and data_obj.inhibitions.issubset(data_food_catalyzed)
            and data_obj.inhibitions):
            self.assertFalse(result_catalyzed)
            
    @given(name=st.text(base_alphabet, min_size=1), warned = st.booleans(), reactants = standard_list(standard_text()), 
           products = standard_list(standard_text()), catalysts = standard_list(standard_text()), inhibitors = standard_list(standard_text()),
           reactant_coefficients = st.lists(st.integers(min_value=1), min_size=listsize, max_size=listsize), product_coefficients = st.lists(st.integers(min_value=1), min_size=listsize, max_size=listsize),
           direction_reaction = st.integers(0, 2), direction_test = st.integers(0, 2), food_reactants = st.lists(st.text(base_alphabet, min_size=1)),food_catalyst = st.lists(st.text(base_alphabet, min_size=1)), food_inhibition = st.lists(st.text(base_alphabet, min_size=1)))
    def test_inhibitions_catalization_food_specific(self, name:str, warned:bool, reactants:list[str], products:list[str], catalysts:list[str]
                              , inhibitors:list[str], reactant_coefficients:list[int], product_coefficients:list[int],
                              direction_reaction, direction_test, food_reactants:list[str], food_catalyst:list[str], food_inhibition:list[str]):
        
        data_obj:Reaction = self.basic_to_reaction(name, warned, reactants, products, catalysts, inhibitors, reactant_coefficients, product_coefficients, direction_reaction)
        data_food_catalyzed = MoleculeType().values_of(food_reactants)
        data_food_catalysts = MoleculeType().values_of(food_catalyst)
        data_food_inhibitors = MoleculeType().values_of(food_inhibition)
        
        data_obj_catalysts = MoleculeType().values_of(data_obj.catalysts.split(","))
        
        match str(direction_test):
            case "0":
                direction_key = "forward"
            case "1":
                direction_key = "reverse"
            case "2":
                direction_key = "both"
        result_catalyzed = data_obj.is_catalyzed_uninhibited_all_reactants(direction_key, food_for_reactants = data_food_catalyzed,
                                                                           food_for_catalysts = data_food_catalysts,
                                                                           food_for_inhibitions = data_food_inhibitors)
        
        if (direction_key in {"forward" or "both"}
            and data_obj.reactants.issubset(data_food_catalyzed) 
            and data_obj_catalysts.issubset(data_food_catalysts)
            and data_obj.inhibitions.isdisjoint(data_food_inhibitors)):
            self.assertTrue(result_catalyzed)
        elif (direction_key in {"forward" or "both"}
            and data_obj.reactants.issubset(data_food_catalyzed) 
            and data_obj_catalysts.issubset(data_food_catalysts)
            and data_obj.inhibitions.isdisjoint(data_food_inhibitors)):
            self.assertTrue(result_catalyzed)
        elif (direction_key in {"forward" or "both"}
            and data_obj.reactants.issubset(data_food_catalyzed) 
            and data_obj_catalysts.issubset(data_food_catalysts)
            and data_obj.inhibitions.issubset(data_food_inhibitors)
            and data_obj.inhibitions):
            self.assertFalse(result_catalyzed)
        elif (direction_key in {"forward" or "both"}
            and data_obj.reactants.issubset(data_food_catalyzed) 
            and data_obj_catalysts.issubset(data_food_catalysts)
            and data_obj.inhibitions.issubset(data_food_inhibitors)
            and data_obj.inhibitions):
            self.assertFalse(result_catalyzed)
     
    @given(name=st.text(base_alphabet, min_size=1), warned = st.booleans(), reactants = standard_list(standard_text()), 
           products = standard_list(standard_text()), catalysts = standard_list(standard_text()), inhibitors = standard_list(standard_text()),
           reactant_coefficients = st.lists(st.integers(min_value=1), min_size=listsize, max_size=listsize), product_coefficients = st.lists(st.integers(min_value=1), min_size=listsize, max_size=listsize),
           direction = st.integers(0, 2), tabbed = st.booleans())
    def parser(self, name:str, warned:bool, reactants:list[str], products:list[str], catalysts:list[str]
                              , inhibitors:list[str], reactant_coefficients:list[int], product_coefficients:list[int],
                              direction:int, tabbed:bool):
        """doesnt test for and in catalysts"""
        """ assume(all([react.isalpha() for react in reactants]))
        assume(all([product.isalpha() for product in products]))
        assume(all([catalyst.isalpha() for catalyst in catalysts]))
        assume(all([inhibitor.isalpha() for inhibitor in inhibitors])) """
        result = self.basic_to_reaction(name, warned, reactants, products, catalysts, inhibitors, reactant_coefficients, product_coefficients, direction)
        if not tabbed:
            test_text = name + "/t:"
            for react in reactants:
                test_text += " " + str(result.reactant_coefficients[MoleculeType().value_of(react)]) + " " + react
            test_text += " ["
            for i, c in enumerate(catalysts):
                if i == 0:
                    test_text += c
                else: test_text += "," + c
            test_text += "] {"
            for i, inh in enumerate(inhibitors):
                if i == 0:
                    test_text += inh
                else: test_text += "," + inh
            test_text += "}"
            match str(direction):
                case "0":
                    test_text += " -> "
                case "1":
                    test_text += " <- "
                case "2":
                    test_text += " <-> "
            for product in products:
                test_text += " " + str(result.product_coefficients[MoleculeType().value_of(product)]) + " " + product
        elif tabbed:
            test_text = name + " /t "
            for react in reactants:
                test_text += " " + str(result.reactant_coefficients[MoleculeType().value_of(react)]) + " " + react
            match str(direction):
                case "0":
                    test_text += " -> "
                case "1":
                    test_text += " <- "
                case "2":
                    test_text += " <-> "
            for product in products:
                test_text += " " + str(result.product_coefficients[MoleculeType().value_of(product)]) + " " + product
            test_text += " /t ["
            for i, c in enumerate(catalysts):
                if i == 0:
                    test_text += c
                else: test_text += "," + c
            test_text += "] /t {"
            for i, inh in enumerate(inhibitors):
                if i == 0:
                    test_text += inh
                else: test_text += "," + inh
            test_text += "}"
        
        test_obj = Reaction().parse_new(test_text, tabbed_format=tabbed)
        
        self.assertEqual(test_obj,result)
            
    def dnf(self):
        
        data_str = ["AB,CD&EF&GH,IJ", "AB,(CD&EF)&GH,IJ", "AB,(CD,EF)&GH,IJ", "(AB,CD)&(EF,GH),IJ"]
        dummy_res_str = [{"AB","CD&EF&GH","IJ"}, {"AB","CD&EF&GH","IJ"}, {"AB","CD&GH","EF&GH","IJ"}, {"AB&EF","AB&GH","CD&EF","CD&GH","IJ"}]
        res_str = [MoleculeType().values_of(list(dummy_res_str[i])) for i in range(0,len(dummy_res_str))]
        test_str = [Reaction("", catalysts=data).get_catalyst_conjunctions() for data in data_str]
        print([[mol.name for mol in s] for s in test_str])
        self.assertEqual(test_str, res_str)
        
    def dnf_elements(self):
        data_str = ["AB,CD&EF&GH,IJ", "AB,(CD&EF)&GH,IJ", "AB,(CD,EF)&GH,IJ", "(AB,CD)&(EF,GH),IJ"]
        dummy_res_str = [{"AB","CD","EF","GH","IJ"}, {"AB","CD","EF","GH","IJ"},{"AB","CD","EF","GH","IJ"},{"AB","CD","EF","GH","IJ"}]
        res_str = [MoleculeType().values_of(list(dummy_res_str[i])) for i in range(0,len(dummy_res_str))]
        test_str = [Reaction("", catalysts=data).get_catalyst_elements() for data in data_str]
        print([[mol.name for mol in s] for s in test_str])
        self.assertEqual(test_str, res_str)
                

class ReactionSystemTests(unittest.TestCase):
    
    def basic_to_Reactionsystem(self,reaction_names:list[str], warned:list[bool], reactants:list[list[str]], products:list[list[str]], catalysts:list[list[str]]
                              , inhibitors:list[list[str]], reaction_coefficients:list[list[int]], product_coefficients:list[list[int]],
                              direction:list[int], foods:list[str], rs_name:str)->ReactionSystem:
        sys_reactions = []
        for i, name in enumerate(reaction_names):
            sys_reactions.append(ReactionTests().basic_to_reaction(reaction_names[i], warned[i], reactants[i], 
                                                             products[i], catalysts[i], inhibitors[i], 
                                                             reaction_coefficients[i], product_coefficients[i], direction[i]))
        sys_foods = MoleculeType().values_of(foods)
        return ReactionSystem(rs_name, reactions=sys_reactions, foods=sys_foods)
    
    @given(reaction_names=standard_list(standard_text()), 
           warned = st.lists(st.booleans(), min_size=listsize, max_size=listsize),
           reactants = standard_list(standard_list(standard_text())), 
           products = standard_list(standard_list(standard_text())), 
           catalysts = standard_list(standard_list(standard_text())), 
           inhibitors = standard_list(standard_list(standard_text())),
           reaction_coefficients = standard_list(standard_list(st.integers(min_value=1))), 
           product_coefficients = standard_list(standard_list(st.integers(min_value=1))),
           direction = standard_list(st.integers(0, 2)),
           foods = standard_list(standard_text()),
           rs_name=standard_text())  
    @settings(max_examples=20)
    def test_copy(self,reaction_names:list[str], warned:list[bool], reactants:list[list[str]], products:list[list[str]], catalysts:list[list[str]]
                              , inhibitors:list[list[str]], reaction_coefficients:list[list[int]], product_coefficients:list[list[int]],
                              direction:list[int], foods:list[str], rs_name:str):
        #st = time.time()
        test_obj = self.basic_to_Reactionsystem(reaction_names, warned, reactants, products, catalysts, inhibitors, reaction_coefficients, product_coefficients, direction, foods, rs_name)
        #et = time.time()
        #print(et-st)
        res_obj = copy.copy(test_obj)
        self.assertEqual(test_obj, res_obj)
    
    @given(reaction_names=standard_list(standard_text()), 
           warned = st.lists(st.booleans(), min_size=listsize, max_size=listsize),
           reactants = standard_list(standard_list(standard_text())), 
           products = standard_list(standard_list(standard_text())), 
           catalysts = standard_list(standard_list(standard_text())), 
           inhibitors = standard_list(standard_list(standard_text())),
           reaction_coefficients = standard_list(standard_list(st.integers(min_value=1))), 
           product_coefficients = standard_list(standard_list(st.integers(min_value=1))),
           direction = standard_list(st.integers(0, 2)),
           foods = standard_list(standard_text()),
           rs_name=standard_text(),
           outside_foods = standard_list(standard_text()))  
    @settings(max_examples=20)
    def test_mentioned_foods(self,reaction_names:list[str], warned:list[bool], reactants:list[list[str]], products:list[list[str]], catalysts:list[list[str]]
                              , inhibitors:list[list[str]], reaction_coefficients:list[list[int]], product_coefficients:list[list[int]],
                              direction:list[int], foods:list[str], rs_name:str, outside_foods:list[str]):
        test_obj = self.basic_to_Reactionsystem(reaction_names, warned, reactants, products, catalysts, inhibitors, reaction_coefficients, product_coefficients, direction, foods, rs_name)
        molecules = []
        for one_type in [reactants, products, inhibitors, catalysts]:
            for reactions in one_type:
                for reaction in reactions:
                    molecules.append(reaction)
        molecules = list(set(molecules).intersection(outside_foods))
        result = MoleculeType().values_of(molecules)
        test_res = test_obj.compute_mentioned_foods(MoleculeType().values_of(molecules))
        """ print("function: ")
        print([mol.name for mol in test_res])
        print("constructed: ")
        print([mol.name for mol in result]) """
        self.assertEqual(result, test_res)
        
class ModelIOTests(unittest.TestCase):
    
    def test_in_is_out(self):
        global application_path
        file_path = "G:/Github/BA-Jan/test_files/inhibitions-1.crs"
        test_path = "G:/Github/BA-Jan/test_results/result.crs"
        res_lines = []
        notation = (ReactionNotation.FULL, ArrowNotation.USES_EQUALS)
        with open(file_path, "r") as f:
            res_lines = f.readlines()
            notation = ReactionNotation.detect_notation(res_lines)
        test_obj = ReactionSystem()
        ModelIO.read(test_obj, file_path, notation[0])
        with open(test_path, "w") as f:
            f.write(ModelIO().write(test_obj, True, notation[0], notation[1]))
            
        #print(ModelIO().write(test_obj, True, ReactionNotation.FULL, ArrowNotation.USES_MINUS),test_path)
        with open(test_path) as file:
            test_lines = file.readlines()
            
        self.assertEqual(res_lines, test_lines)
    
            
        
if __name__ == '__main__':
    unittest.main()
    print(sys.path)


from __future__ import annotations
from enum import StrEnum

from .ArrowNotation import ArrowNotation


class ReactionNotation(StrEnum):

    __order__ = "FULL SPARSE TABBED"

    FULL = "FULL"
    SPARSE = "SPARSE"
    TABBED = "TABBED"

    def detect_notation(lines: list[str]) -> tuple[ReactionNotation, ArrowNotation] | None:
        """detects the minimum line format given a list of lines.

        Args:
            lines (list[str]): lines decribing food and reactions

        Returns:
            tuple[ReactionNotation, ArrowNotation] | None: the minimum reaction notation and arrow notation
            
        Prefers FULL and SPARSE to TABBED
        """
        arrows_use_equals = False
        arrows_use_minus = False
        contains_tabs = False
        contains_square_brackets = False
        contains_commas = False

        for line in lines:
            if not line.startswith(("#", "Food:", "F:")):
                if "\t" in line:
                    contains_tabs = True
                if "[" in line or "]" in line:
                    contains_square_brackets = True
                if "=>" in line or "<=" in line:
                    arrows_use_equals = True
                if "->" in line or "<-" in line:
                    arrows_use_minus = True
            elif not line.startswith("#"):
                if "," in line:
                    contains_commas = True
        if arrows_use_minus or arrows_use_equals:
            if contains_square_brackets:
                if contains_commas:
                    return (ReactionNotation.FULL, ArrowNotation.USES_EQUALS if arrows_use_equals else ArrowNotation.USES_MINUS)
                else:
                    return (ReactionNotation.SPARSE, ArrowNotation.USES_EQUALS if arrows_use_equals else ArrowNotation.USES_MINUS)
            elif contains_tabs:
                return (ReactionNotation.TABBED, ArrowNotation.USES_EQUALS if arrows_use_equals else ArrowNotation.USES_MINUS)
        return None



from __future__ import annotations
import sys
import os
sys.path.insert(0, os.path.abspath(
    os.path.join(os.path.dirname(__file__), '.')))
from model.ReactionSystem import ReactionSystem
from enum import StrEnum
from tqdm import tqdm
import networkx as nx
import matplotlib.pyplot as plt



class edge_types(StrEnum):

    REACTANT = "reactant"
    PRODUCT = "product"
    INHIBITOR = "inhibitor"
    CATALYST = "catalyst"
    CATALYST_CONJUNCTION = "catalyst_conjuntion"


class node_types(StrEnum):

    REACTION = "reaction"
    MOLECULE = "molecule"
    CATALYST_CONJUNCTION = "catalyst_conjunction"


SUPPORTED_GRAPH_OUTPUT_FILE_FORMATS = {".gml", ".graphml"}

ARROW_SHAPE_DICT = {edge_types.INHIBITOR:'T',
                    edge_types.CATALYST:"Arrow",
                   edge_types.CATALYST_CONJUNCTION:"Arrow",
                   edge_types.REACTANT:"Arrow",
                   edge_types.PRODUCT:"Arrow"}
NODE_SHAPE_DICT = {node_types.MOLECULE:'ELLIPSE', 
                   node_types.REACTION:'TRIANGLE', 
                   node_types.CATALYST_CONJUNCTION:'OCTAGON'}

EDGE_COLOR_DICT = {edge_types.INHIBITOR:"#FF0000",
                   edge_types.CATALYST:"#00FF00",
                   edge_types.CATALYST_CONJUNCTION:"#00FF00",
                   edge_types.REACTANT:"#000000",
                   edge_types.PRODUCT:"#000000"}
NODE_COLOR_DICT = {node_types.MOLECULE:'#d3d3d3', 
                   node_types.REACTION:'#d3d3d3', 
                   node_types.CATALYST_CONJUNCTION:'#d3d3d3'}

def write(reaction_systems: list[ReactionSystem], 
          filename: str, output_format:str) -> None:
    """Generate '.gml' files from a list of reactions systems.

    Args:
        reaction_systems (list[ReactionSystem]): list of reaction systems to be written
        filename (str): path of output file. If reaction_systems contains more than 1 reaction system
                        this path will be a directory with 'filename' + index as the new filenames.
        algorithm (AlgorithmBase, optional): Only used for Message Generation if a reaction system is empty. Defaults to None.
    """
    if filename == "stdout":
        tqdm.write("stdout is not currently available for graph format")
        return None
    
    if isinstance(reaction_systems, list):
        if len(reaction_systems) > 1:
            output = os.path.split(os.path.abspath(filename))
            output_directory = os.path.join(output[0], output[1].split(".")[0])
            output_file = output[1]
            os.makedirs(os.path.dirname(output_directory), exist_ok=True)

    for i, rs in enumerate(reaction_systems):
        graph = parse_rs_to_graph(rs)
        if len(reaction_systems) > 1:
            output_file = ".".join(
                [output_file.split(".")[0] + str(i), output_file.split(".")[1]])
            filename = os.path.join(output_directory, output_file)
        if ".gml" == output_format:
            tqdm.write("Writing file: " + filename)
            tqdm.write(rs.get_header_line())
            nx.write_gml(graph, filename)
        elif ".graphml" in filename:
            res_node_dict = {} #node:{attr:..., attr:...}
            node_keys_set = set()
            for label in list(graph.nodes()):
                node = graph.nodes()[label]
                single_node_dict = {}
                for attr in ['graphics', 'att']:
                    flat_dict = flatten_dict_dpth_2(node[attr], attr)
                    node_keys_set.update(set(flat_dict.keys()))
                    single_node_dict.update(flat_dict)
                    node.pop(attr)
                res_node_dict.update({label:single_node_dict})
            res_edge_dict = {}
            edge_keys_set = set()
            for label in list(graph.edges()):
                edge = graph.edges()[label]
                single_edge_dict = {}
                for attr in ['graphics', 'att']:
                    flat_dict = flatten_dict_dpth_2(edge[attr], attr)
                    edge_keys_set.update(set(flat_dict.keys()))
                    single_edge_dict.update(flat_dict)
                    edge.pop(attr)
                res_edge_dict.update({label:single_edge_dict})
            for key in node_keys_set:
                nx.set_node_attributes(graph, "", name=key)
            for key in edge_keys_set:
                nx.set_edge_attributes(graph, "", name=key)
            for label in list(graph.nodes()):
                node = graph.nodes()[label]
                for key in node_keys_set:
                    try:
                        graph.nodes()[label][key] = res_node_dict[label][key]
                    except: pass
                if node['att#node_type'] == node_types.REACTION:
                    node.pop('att#Food')
            for label in list(graph.edges()):
                edge = graph.edges()[label]
                for key in edge_keys_set:
                    graph.edges()[label][key] = res_edge_dict[label][key]
            nodes = list(graph.nodes(data=True))
            edges = list(graph.edges(data=True))
            nx.write_graphml(graph, filename)
        else:
            tqdm.write("File format not recognized." +
                       " Assumed .gml.")
            tqdm.write(rs.get_header_line())
            nx.write_gml(graph, filename)

def flatten_dict_dpth_2(attr:dict, key:str, sep:str = "#") -> dict[str, str]:
    res = {}
    for deeper_key in list(attr.keys()):
        val = attr[deeper_key]
        res.update({key + sep + deeper_key:val})
    
    return res

def parse_rs_to_graph(reaction_system: ReactionSystem) -> nx.DiGraph:
    """parses a reaction system to a networkx DiGraph object.

    Args:
        reaction_system (ReactionSystem): the rs to be parsed

    Returns:
        nx.DiGraph: An equivalent graph
        
    The resulting graph has nodes for:
     - Molecules (ellipses, white fill)
     - Reactions (triangles, black fill)
     - Catalyst Conjunctions (octagons, white fill)
    The edges are the connecting interactions/relations. So:
     - Reactants: point from molecules to reactions (Arrow, black)
     - Products: point away from reactions to molecules (Arrow, black)
     - Catalysts: point from molecules or catalyst conjunctions to reactions (Arrow, green)
     - Catalyst Conjunctions: point from molecules to catalyst conjunctions (Arrow, green)
     - Inhibitions: point from molecules to reactions (T, red)
    
    The graphics attributes are saved under: 
    Nodes:
     - ['graphics'] ['NodeShape'] (str): Shape of the node
     - ['graphics'] ['fill'] (str): Hexcode of the node fill color
    Edges:
     - ['graphics'] ['ArrowShape'] (str): Shape of the Arrow
     - ['graphics'] ['color'] (str): Hexcode of the edge color
    
    Nodes have the additional attributes:
     - ['att'] ['node_type'] (node_types) Identifies their node type according to the list above
     - ['att'] ['Food'] (bool) Identifies if a node is part of the food set 
     
    Edges have the additional attributes:
     - ['att'] ['edge_type'] (edge_types) Identifies their edge type according to the list above
     - ['weight'] (float|str|None) if a coefficient for a reactant or product edge is present, it is added here.
    """
    graph = nx.DiGraph(name=reaction_system.name)
    molecule_nodes = reaction_system.get_mentioned_molecules()
    molecule_nodes = [(node.name,
                       {"graphics": {"NodeShape": NODE_SHAPE_DICT[node_types.MOLECULE],
                                     "fill": NODE_COLOR_DICT[node_types.MOLECULE]},
                        "att": {"node_type": node_types.MOLECULE.value,
                                "Food": False}})
                      for node in molecule_nodes]
    graph.add_nodes_from(molecule_nodes)
    for food in reaction_system.foods:
        graph.nodes[food.name]["att"]["Food"] = True
    reaction_nodes = [(reaction.name,
                       {"graphics": {"NodeShape": NODE_SHAPE_DICT[node_types.REACTION],
                                     "fill": NODE_COLOR_DICT[node_types.REACTION]},
                        "att": {"node_type": node_types.REACTION.value}})
                      for reaction in reaction_system.reactions]
    graph.add_nodes_from(reaction_nodes)

    for reaction in reaction_system.reactions:
        for i, reactant in enumerate(reaction.reactants):
            if reaction.reactant_coefficients:
                coefficient = reaction.reactant_coefficients[i]
            else:
                coefficient = None
            match reaction.direction:
                case "forward":
                    parse_edge(graph, reactant.name, reaction.name,
                               edge_types.REACTANT, coefficient)
                case "reverse":
                    parse_edge(graph, reaction.name, reactant.name,
                               edge_types.PRODUCT, coefficient)
                case "both":
                    parse_edge(graph, reactant.name, reaction.name,
                               edge_types.REACTANT, coefficient)
                    parse_edge(graph, reaction.name, reactant.name,
                               edge_types.PRODUCT, coefficient)
        for i, product in enumerate(reaction.products):
            if reaction.product_coefficients:
                coefficient = reaction.product_coefficients[i]
            else:
                coefficient = None
            match reaction.direction:
                case "reverse":
                    parse_edge(graph, product.name, reaction.name,
                               edge_types.REACTANT, coefficient)
                case "forward":
                    parse_edge(graph, reaction.name, product.name,
                               edge_types.PRODUCT, coefficient)
                case "both":
                    parse_edge(graph, product.name, reaction.name,
                               edge_types.REACTANT, coefficient)
                    parse_edge(graph, reaction.name, product.name,
                               edge_types.PRODUCT, coefficient)
        for inhibitor in reaction.inhibitions:
            parse_edge(graph, inhibitor.name, reaction.name,
                       edge_types.INHIBITOR)
        for catalyst in [cata.name for cata in reaction.get_catalyst_conjunctions()]:
            if "&" in catalyst:
                catalyst_node = [(catalyst, {"graphics": {"NodeShape": NODE_SHAPE_DICT[node_types.CATALYST_CONJUNCTION],
                                                          "fill": NODE_COLOR_DICT[node_types.CATALYST_CONJUNCTION]},
                                             "att": {"node_type":
                                                     node_types.CATALYST_CONJUNCTION.value}})]
                graph.add_nodes_from(catalyst_node)
                parse_edge(graph, catalyst, reaction.name, edge_types.CATALYST)
                catalyst_elements = catalyst.split("&")
                for catalyst_element in catalyst_elements:
                    parse_edge(graph, catalyst_element, catalyst,
                               edge_types.CATALYST_CONJUNCTION)
            else:
                parse_edge(graph, catalyst, reaction.name, edge_types.CATALYST)

    return graph

def print_graph(graph:nx.DiGraph) -> None:
    
    pos = nx.spring_layout(graph)
    nodes = graph.nodes(data=True)
    edges = graph.edges(data=True)
    node_shape_dict = {'circle':[], 
                       'triangle':[], 
                       'hexagon':[]}
    for node in nodes:
        if node[1]['graphics']['NodeShape'] == 'ELLIPSE':
            node_shape_dict["circle"].append(node[0])
        elif node[1]['graphics']['NodeShape'] == 'TRIANGLE':
            node_shape_dict["triangle"].append(node[0])
        elif node[1]['graphics']['NodeShape'] == 'OCTAGON':
            node_shape_dict["hexagon"].append(node[0])
        else:
            tqdm.write("Node: " + node[0] + "could not be recognized as molecule, reaction or catalyst conjunction")

    for edge in edges:
        if edge[2]['graphics']['ArrowShape'] == 'Arrow':
            style = '->'
        else:
            style = '|-|'
        try:
            weight = edge[2]['weight']
        except KeyError:
            weight = 1
        nx.draw_networkx_edges(graph, 
                               pos, 
                               edgelist=[edge],
                               width = weight,
                               edge_color=edge[2]['graphics']['color'],
                               arrowstyle=style)
    
    nx.draw_networkx_nodes(graph, pos, nodelist=node_shape_dict["circle"], 
                           node_color=NODE_COLOR_DICT[node_types.MOLECULE], node_shape='o')
    nx.draw_networkx_nodes(graph, pos, nodelist=node_shape_dict["triangle"], 
                           node_color=NODE_COLOR_DICT[node_types.REACTION], node_shape='^')
    nx.draw_networkx_nodes(graph, pos, nodelist=node_shape_dict["hexagon"], 
                           node_color=NODE_COLOR_DICT[node_types.CATALYST_CONJUNCTION], node_shape='h')
    
    nx.draw_networkx_labels(graph, pos)
    
    plt.show()
    
def print_rs_as_graph(reaction_system:ReactionSystem) -> None:
    
    graph = parse_rs_to_graph(reaction_system)
    print_graph(graph)

def parse_edge(graph: nx.DiGraph,
               u: str,
               v: str,
               edge_type: edge_types,
               coefficient: str | None = None) -> None:
    """Adds one edge in the appropriate format given edge type.

    Args:
        graph (nx.DiGraph): Graph the edge should be added to
        u (str): name of the starting node of the edge
        v (str): name of the end node of the edge
        edge_type (edge_types): Format the edge should take
        coefficient (str | None, optional): reactant-/product coefficient. Is converted to weight. Defaults to None.
    """    
    if edge_type == edge_types.INHIBITOR:
        color = "#FF0000"
        arrow = "T"
    elif edge_type in [edge_types.CATALYST, edge_types.CATALYST_CONJUNCTION]:
        color = "#00FF00"
        arrow = "Arrow"
    elif edge_type in [edge_types.REACTANT, edge_types.PRODUCT]:
        color = "#000000"
        arrow = "Arrow"
    else:
        tqdm.write("Edge type: " + edge_type + " isn't recognized." +
                   "REACTANT properties are used.")
        color = "#000000"
        arrow = "Arrow"

    graph.add_edge(u, v, graphics={"color": color,
                                   "ArrowShape": arrow},
                   att={"edge_type": edge_type.value})
    if coefficient:
        try:
            graph[u][v]["weight"] = float(coefficient)
        except ValueError:
            graph[u][v]["weight"] = coefficient



from __future__ import annotations
import re
from settings.ArrowNotation import ArrowNotation
from settings.ReactionNotation import ReactionNotation
from model.ReactionSystem import ReactionSystem, MoleculeType
from model.Reaction import Reaction, FORMAL_FOOD
from tqdm import tqdm
import sys
import os
sys.path.insert(0, os.path.abspath(
    os.path.join(os.path.dirname(__file__), '..')))

SUPPORTED_OUTPUT_FILE_FORMATS = set([".crs"])

SUPPORTED_INPUT_FILE_FORMATS = set([".crs"])


class ModelIO:

    def parse_food(a_line: str) -> set[MoleculeType]:
        """Determines if a line is a food line and parses to a list of molecules.

        Args:
            a_line (str): line of text

        Returns:
            list[MoleculeType]: list of food molecules or empty list
        """
        a_line = re.sub(" +", " ", a_line.replace(",", " "))
        if a_line.startswith("Food:"):
            if len(a_line) > len("Food:"):
                a_line = a_line.removeprefix("Food:").strip()
            else:
                a_line = ""
        elif a_line.startswith("Food"):
            if len(a_line) > len("Food"):
                a_line = a_line.removeprefix("Food").strip()
            else:
                a_line = ""
        elif a_line.startswith("F:"):
            if len(a_line) > len("F:"):
                a_line = a_line.removeprefix("F:").strip()
            else:
                a_line = ""

        return MoleculeType().values_of(a_line.split())

    def read(reaction_system: ReactionSystem, filename: str, reaction_notation: ReactionNotation) -> str:
        """Prases file into a reaction system

        Args:
            reaction_system (ReactionSystem): ReactionSystem to be filled
            filename (str): path to file
            reaction_notation (ReactionNotation): Reaction Format. i.e. TABBED

        Raises:
            IOError: Multiple Reactions with the same name detected

        Returns:
            str: leading comments of the file
        """
        reaction_names: set[str] = set()
        in_leading_comments: bool = True
        leading_comments: list[str] = []

        with open(filename, "r") as f:
            lines = f.readlines()
            for i, line in enumerate(lines):
                if not line.startswith("#"):
                    in_leading_comments = False
                    line = line.strip()
                    if len(line) > 0:
                        try:
                            arrow_bool = not any(
                                [arrow in line for arrow in ["->", "=>", "<-", "<="]])
                            if (line.startswith("Food:")
                                or (line.startswith("F:")
                                    and arrow_bool)):
                                reaction_system.foods.update(
                                    ModelIO.parse_food(line))
                            else:
                                reaction = Reaction().parse_new(line, reaction_notation)
                                if reaction.name in reaction_names:
                                    raise IOError(
                                        "Multiple reactions have the same name:\t"
                                        + str(reaction.name))
                                reaction_system.reactions.append(reaction)
                                reaction_names.add(reaction.name)
                                if (FORMAL_FOOD.name in reaction.catalysts 
                                    and not FORMAL_FOOD in reaction_system.foods):
                                    reaction_system.foods.add(FORMAL_FOOD)
                        except IOError as e:
                            msg = e.args[0]
                            raise IOError(msg, i)
                elif in_leading_comments:
                    leading_comments.append(line)

        return "\n".join(leading_comments)

    def write(self,
              reaction_system: ReactionSystem,
              include_food: bool,
              reaction_notation: ReactionNotation,
              arrow_notation: ArrowNotation,
              food_first: bool = True) -> str:
        """Parses a reaction system to a str in the format of a '.crs' file.

        Args:
            reaction_system (ReactionSystem): reaction system to be parsed
            include_food (bool): should food be icluded in the file
            reaction_notation (ReactionNotation): Determines reaction notation format i.e. TABBED
            arrow_notation (ArrowNotation): Determines arrow format i.e. USES_EQUALS
            food_first (bool, optional): Should the food be at the beginning or end of the file. Defaults to True.

        Returns:
            str: str format of reaction system
        """        
        if not reaction_system.reactions:
            tqdm.write("The resulting reaction system has no reactions")
            return "The resulting reaction system has no reactions"
        res = ""
        if food_first and include_food:
            res += "Food: " + ModelIO().get_food_str(reaction_system, reaction_notation) + "\n\n"

        for reaction in reaction_system.reactions:
            res += ModelIO().get_reaction_str(reaction, reaction_notation, arrow_notation) + "\n"

        if not food_first and include_food:
            res += "Food: " + ModelIO().get_food_str(reaction_system, reaction_notation) + "\n\n"
        return res

    def get_food_str(self,
                     reaction_system: ReactionSystem,
                     reaction_notation: ReactionNotation) -> str:
        try:
            foods = reaction_system.foods
            foods = [food.name for food in reaction_system.foods]
            foods.sort()
            if reaction_notation == ReactionNotation.FULL:
                return ", ".join(foods)
            else:
                return " ".join(foods)
        except IOError:
            return ""

    def get_reaction_str(self,
                         reaction: Reaction,
                         reaction_notation: ReactionNotation,
                         arrow_notation: ArrowNotation) -> str:
        """str representation of a reaction for a '.crs' file

        Args:
            reaction (Reaction): reacton to be parsed to str
            reaction_notation (ReactionNotation): Determines reaction notation format i.e. TABBED
            arrow_notation (ArrowNotation): Determines arrow format i.e. USES_EQUALS

        Returns:
            str: str representation of a reaction
        """        
        res = ""
        sep = " "
        arrow = ""
        match reaction.direction:
            case "forward":
                arrow = " => " if arrow_notation == ArrowNotation.USES_EQUALS else " -> "
            case "reverse":
                arrow = " <= " if arrow_notation == ArrowNotation.USES_EQUALS else " <- "
            case "both":
                arrow = " <=> " if arrow_notation == ArrowNotation.USES_EQUALS else " <-> "

        if reaction_notation == ReactionNotation.FULL:
            sep = ","
        if reaction_notation == ReactionNotation.TABBED:
            res = reaction.name + "\t"
            reactants_and_coefficients = []
            for reactant in reaction.reactants:
                try:
                    reactants_and_coefficients.append(
                        str(reaction.reactant_coefficients[reactant.name])
                        + " " +
                        reactant.name)
                except:
                    reactants_and_coefficients.append(reactant.name)
            reactants_and_coefficients.sort()
            res += " + ".join(reactants_and_coefficients)
            res += arrow
            products_and_coefficients = []
            for product in reaction.products:
                try:
                    products_and_coefficients.append(
                        str(reaction.product_coefficients[product.name])
                        + " " +
                        product.name)
                except:
                    products_and_coefficients.append(product.name)
            products_and_coefficients.sort()
            res += " + ".join(products_and_coefficients)
            res += "\t"
            if reaction.catalysts != "":
                res += "[" + reaction.catalysts + "]"
            if reaction.inhibitions:
                res += "\t{"
                inh_list = list(reaction.inhibitions)
                inh_list.sort()
                for i, inh in enumerate(inh_list):
                    if i == 0:
                        res += inh.name
                    else:
                        res += "," + inh.name
                res += "}"
            return res
        else:
            res = reaction.name + " : "
            reactants_and_coefficients = []
            for reactant in reaction.reactants:
                try:
                    reactants_and_coefficients.append(
                        str(reaction.reactant_coefficients[reactant.name])
                        + " "
                        + reactant.name)
                except:
                    reactants_and_coefficients.append(reactant.name)
            reactants_and_coefficients.sort()
            res += " + ".join(reactants_and_coefficients)
            res += " "
            if reaction.catalysts != "":
                res += "[" + reaction.catalysts + "]"
            if reaction.inhibitions:
                res += " {"
                inh_list = list(reaction.inhibitions)
                inh_list.sort()
                for i, inh in enumerate(inh_list):
                    if i == 0:
                        res += str(inh.name)
                    else:
                        res += sep + inh.name
                res += "}"
            res += arrow
            products_and_coefficients = []
            for product in reaction.products:
                try:
                    products_and_coefficients.append(
                        str(reaction.product_coefficients[product.name])
                        + " "
                        + product.name)
                except:
                    products_and_coefficients.append(product.name)
            products_and_coefficients.sort()
            res += " + ".join(products_and_coefficients)
            return res



from __future__ import annotations
from model.MoleculeType import MoleculeType
from copy import copy, deepcopy
from model.DisjunctiveNormalForm import compute
from tqdm import tqdm
import re
import sys
import os
sys.path.insert(0, os.path.abspath(
    os.path.join(os.path.dirname(__file__), '.')))

FORMAL_FOOD = MoleculeType().value_of(name="$")


class Reaction:
    """A reaction.
    """

    """ @property
    def catalysts(self):
        return self._catalysts

    @catalysts.setter
    def catalysts(self, value: str):
        self._catalysts = value
        buffer = self.get_catalyst_conjunctions()
        self._catalyst_conjunctions = buffer """

    def __deepcopy__(self, memo) -> Reaction:
        id_self = id(self)
        _copy = memo.get(id_self)
        if _copy == None:
            _copy = Reaction(deepcopy(self.name, memo),
                             warned_about_suppressing_coefficients=deepcopy(
                                 self.warned_about_suppressing_coefficients, memo),
                             reactants=deepcopy(self.reactants, memo),
                             products=deepcopy(
                                 self.products, memo),
                             catalysts=deepcopy(self.catalysts, memo),
                             catalyst_conjunctions=deepcopy(
                                 self.catalyst_conjunctions),
                             inhibitions=deepcopy(self.inhibitions, memo),
                             reactant_coefficients=deepcopy(
                                 self.reactant_coefficients, memo),
                             product_coefficients=deepcopy(
                                 self.product_coefficients, memo),
                             direction=deepcopy(self.direction, memo))
            memo[id_self] = _copy
        return _copy

    def __init__(self, name: str = None, **kwargs) -> Reaction:
        """A Reaction

        Args:
            name (str, optional): name of the reaction. Defaults to None.

        KwArgs:
            reactants (set[MoleculeType]): reactants used in reaction. Defaults to []
            products (list[MoleculeType]): products produced by reaction. Defaults to []
            catalysts (str): catalysts conjunction for reaction. Defaults to ""
            inhibitions (list[MoleculeType]): inhibitions of reaction. Defaults to []
            reactant_coefficients (dict[MoleculeType:float]): reactants mapped to their coefficients. Defaults to {}
            product_coefficients (dict[MoleculeType:float]): products mapped to their coefficients. Defaults to {}
            direction (str): direction the reaction flows in. Defaults to "forward".

        Returns:
            Reaction: the resutling reaction
        """
        self.DIRECTION = {"forward": 'forward',
                          "reverse": 'reverse', "both": 'both'}

        self.warned_about_suppressing_coefficients = False if "warned_about_suppressing_coefficients" not in kwargs else kwargs[
            "warned_about_suppressing_coefficients"]
        self.name = name
        self.reactants: set[MoleculeType] = set(
        ) if "reactants" not in kwargs else kwargs["reactants"]
        self.products: set[MoleculeType] = set(
        ) if "products" not in kwargs else kwargs["products"]
        self.catalysts: str = "" if "catalysts" not in kwargs else kwargs["catalysts"]
        self.catalyst_conjunctions: set[MoleculeType] = set(
        ) if "catalyst_conjunctions" not in kwargs else kwargs["catalyst_conjunctions"]
        self.inhibitions: set[MoleculeType] = set(
        ) if "inhibitions" not in kwargs else kwargs["inhibitions"]
        self.reactant_coefficients = {
        } if "reactant_coefficients" not in kwargs else kwargs["reactant_coefficients"]
        self.product_coefficients = {
        } if "product_coefficients" not in kwargs else kwargs["product_coefficients"]
        self.direction: str = self.DIRECTION["forward"] if "direction" not in kwargs else kwargs["direction"]

    def is_catalyzed_uninhibited_all_reactants(self, direction: str, **kwargs) -> bool:
        """Checks is reaction is uninhibited, catalyzed and has all reactants.

        Args:
            direction (str): direction of the reaction

        KwArgs:\n
        food (set[MoleculeType]): Molecules in food

        food_for_reactants(set[MoleculeType]): Molecules available as reactants
        food_for_catalysts(set[MoleculeType]): Molecules available as catalysts
        food_for_inhibitions(set[MoleculeType]): Molecules available as inhibitions

        Returns:
            bool: True if uninhibited, catalyzed and all reactants present.
                  If any of these are false returns False. 

        Needs either "food"- or all 3 "food_for_..."-kwargs.
        """
        if "food" in kwargs:
            food_set = kwargs["food"]
            return ((direction in {"forward", "both"} and self.reactants.issubset(food_set)
                     or direction in {"reverse", "both"} and self.products.issubset(food_set))
                    and (len(self.catalysts) == 0
                         or any(MoleculeType().values_of(conjunction.name.split("&")).issubset(food_set) for conjunction in self.get_catalyst_conjunctions()))
                    and (len(self.inhibitions) == 0
                         or food_set.isdisjoint(self.inhibitions)))
        else:
            reactant_set = set(
            ) if not "food_for_reactants" in kwargs else kwargs["food_for_reactants"]
            catalyst_set = set(
            ) if not "food_for_catalysts" in kwargs else kwargs["food_for_catalysts"]
            inhibition_set = set(
            ) if not "food_for_inhibitions" in kwargs else kwargs["food_for_inhibitions"]
            return ((direction in {"forward", "both"} and self.reactants.issubset(reactant_set)
                     or direction in {"reverse", "both"} and self.products.issubset(reactant_set))
                    and (len(self.catalysts) == 0
                         or any(MoleculeType().values_of(conjunction.name.split("&")).issubset(catalyst_set) for conjunction in self.get_catalyst_conjunctions()))
                    and (len(self.inhibitions) == 0
                         or not bool(inhibition_set & self.inhibitions)))

    def is_all_reactants(self, food: set[MoleculeType], direction: str) -> bool:
        return (direction in {"forward", "both"}
                and (self.reactants.issubset(food)
                     or direction in {"reverse", "both"})
                and self.products.issubset(food))

    def __eq__(self, other: Reaction | None) -> bool:
        if not (isinstance(other, Reaction)):
            return False
        if self.name != other.name:
            return False
        if self.reactants != other.reactants:
            return False
        if self.products != other.products:
            return False
        if self.catalysts != other.catalysts:
            return False
        if self.inhibitions != other.inhibitions:
            return False
        if self.reactant_coefficients != other.reactant_coefficients:
            return False
        if self.product_coefficients != other.product_coefficients:
            return False
        if self.direction != other.direction:
            return False

        return True

    def __lt__(self, other: Reaction) -> bool:
        return hash(self) < hash(other)

    def __hash__(self) -> int:
        return hash(self.name)

    def parse_new(self, line: str, tabbed_format: bool) -> Reaction:
        """Parses a formatted line to a reaction.

        Args:
            line (str): line describing a reactionin certain formats.
            tabbed_format (bool): true if the format is the tabbed format

        Returns:
            Reaction: reaction represented in line

        Description:
            The accepted formats for the line are a FULL, SPARSE and TABBED format.
            The arrows represent the direction of the reaction with:
                '->' as 'forward'
                '<-' as 'reverse'
                '<->' as 'both'
            They can use '-' or '='.
            The reaction formats can generally be represented by:
            FULL:
            name : [coefficient] reactant ... '[' catalyst ...']'  ['{' inhibitor ... '}'] -> [coefficient] product ...
            or
            name : [coefficient] reactant +  ... '[' catalyst ...']'  ['{' inhibitor ... '}'] -> [coefficient] product + ...
            SPARSE:
            name: [coefficient] reactant ... '[' catalyst ...']'  ['{' inhibitor ... '}'] -> [coefficient] product ...
            or
            name: [coefficient] reactant+ ... '[' catalyst ...']'  ['{' inhibitor ... '}'] -> [coefficient] product+ ...
            TABBED:
            name \\t [coefficient] reactant ...  -> [coefficient] product ... \\t '[' catalyst ...']' \\t ['{' inhibitor ... '}']
            or
            name \\t [coefficient] reactant+ ... -> [coefficient] product+ ... \\t '[' catalyst ...']' \\t ['{' inhibitor ... '}']
        """
        if line.strip() == "":
            return None

        coefficient_bool = False
        line = line.replace("->", "=>")
        line = line.replace("<-", "<=")
        arrow = "=>"
        direction = self.DIRECTION["forward"]
        if "=>" in line:
            direction = self.DIRECTION["forward"]
            arrow = "=>"
        elif "<=" in line:
            direction = self.DIRECTION["reverse"]
            arrow = "<="
        if "<=>" in line:
            direction = self.DIRECTION["both"]
            arrow = "<=>"

        tabs = line.count("\t")
        tabbed_format = tabs > 1
        if tabbed_format:
            tokens: list[str] = line.split("\t")
            if len(tokens) == 2:
                tokens.append("")
            if len(tokens) == 3:
                tokens.append("")
            re_pro = tokens[1].split(arrow)
            tokens[1] = re_pro[0]
            tokens.append(re_pro[1])
        elif len(line.split("\t")) < 2:
            skipped: list[bool] = []
            for sep in [":", "[", "{", arrow]:
                if sep not in line:
                    skipped.append(True)
                else:
                    skipped.append(False)
                line = line.replace(sep, "\t")
            tokens: list[str] = line.split("\t")
            for i, sep in enumerate(skipped):
                if sep:
                    tokens.insert(i+1, "")
        else:
            tqdm.write("Line " + line + " could not be parsed as the " +
                       "parser was unable to recognize the used format.")
            return Reaction()
        for i, token in enumerate(tokens):
            if tabbed_format:
                token = token.replace("[", "")
                token = token.replace("{", "")
            token = token.replace("]", "")
            token = token.replace("}", "")
            tokens[i] = token.strip()

        token_dict: dict[list | str] = {"r_name": tokens[0],
                                        "reactants": tokens[1],
                                        "reactant_coefficients": {},
                                        "catalysts": tokens[2],
                                        "inhibitions": tokens[3],
                                        "products": tokens[4],
                                        "product_coefficients": {}}
        is_all_numeric = True
        for side in ["reactants", "products"]:
            if "+" in token_dict[side]:
                plus_bool = True
                token_dict[side] = token_dict[side].split("+")
            else:
                plus_bool = False
                token_dict[side] = token_dict[side].split()
            if is_all_numeric:
                is_all_numeric = all([r.replace(".", "").replace(",", "")
                                      .replace(" ", "").isdigit()
                                      for r in token_dict[side]])
        for side in ["reactants", "products"]:
            numeric_buffer = [False, False]
            remove_list = []
            for i, r in enumerate(token_dict[side]):
                r = r.strip()
                if not is_all_numeric:
                    if plus_bool:
                        if " " in r:
                            cache = r.split()
                            if (cache[0].isnumeric() or
                                cache[0].replace(".", "", 1)
                                    .replace(",", "", 1).isdigit()):
                                token_dict[side[:-1] +
                                           "_coefficients"].update({cache[1]: cache[0]})
                                r = cache[1]
                            elif (cache[1].isnumeric() or
                                  cache[1].replace(".", "", 1)
                                  .replace(",", "", 1).isdigit()):
                                token_dict[side[:-1] +
                                           "_coefficients"].append({cache[0]: cache[1]})
                                r = cache[0]
                            else:
                                tqdm.write("There was an unexpected whitespace in reaction: " +
                                           token_dict["r_name"] + " in reactant " +
                                           r)
                    else:
                        numeric_buffer[1] = (r.isnumeric() or r.replace(".", "", 1)
                                             .replace(",", "", 1).isdigit())
                        if numeric_buffer[0] and not numeric_buffer[1]:
                            token_dict[side[:-1] +
                                       "_coefficients"].update({r: token_dict[side][i-1]})
                            remove_list.append(i-1)
                        numeric_buffer[0] = numeric_buffer[1]
                else:
                    if " " in r:
                        if not coefficient_bool:
                            tqdm.write("Coefficients are illegal if all"
                                       + " molecules are numeric. Coeff is "
                                       + "assumed to be first number."
                                       + "\nThe first issue occured at reaction: \n"
                                       + token_dict["r_name"] + " in molecule " + r)
                        r = r.split()[1]

                token_dict[side][i] = r
            for index in reversed(remove_list):
                token_dict[side].pop(index)
        token_dict["inhibitions"] = (token_dict["inhibitions"]
                                     .replace(",", " ").split(" ")
                                     if token_dict["inhibitions"] != "" else [])

        catalysts = token_dict["catalysts"]
        if catalysts == "":
            pass
        else:
            catalysts = self.uniform_logic_notation(catalysts)
        token_dict["catalysts"] = catalysts

        res_reaction = Reaction(token_dict["r_name"],
                                inhibitions=MoleculeType().values_of(
                                    token_dict["inhibitions"]),
                                reactants=MoleculeType().values_of(
                                    token_dict["reactants"]),
                                products=MoleculeType().values_of(
                                    token_dict["products"]),
                                catalysts=token_dict["catalysts"],
                                reactant_coefficients=token_dict["reactant_coefficients"],
                                product_coefficients=token_dict["product_coefficients"],
                                direction=direction)
        return res_reaction

    def get_catalyst_conjunctions(self) -> set[MoleculeType]:
        """
        returns conjunctions of catalysts as a set.
        One element of this set is one permutation of elements that can catalyze the reaction.
        The elements within one element of the set are divided by an "&".

        Returns:
            set[MoleculeType]: separated catalyst conjunctions
        """
        dnf = compute(self.catalysts)
        parts = dnf.split(",")
        return MoleculeType().values_of(parts)

    def get_catalyst_elements(self) -> set[MoleculeType]:
        '''
        returns all catalyst elements for this reaction as a set, not considering associations between catalysts.
        '''
        toplevel_conjuncitons = self.get_catalyst_conjunctions()
        all_elements = set()
        for conjunction in toplevel_conjuncitons:
            all_elements.update(MoleculeType().values_of(
                conjunction.name.split("&")))
        return all_elements

    def any_as_forward(self) -> list[Reaction]:
        """Turns this reaction into a list of equivalent reactions with direction 'forward'

        Returns:
            list[Reaction]: either one or two reactions that, together, are equivalent to self
        """
        match self.direction:
            case "forward":
                return [deepcopy(self)]
            case "reverse":
                reverse = deepcopy(self)
                reverse.swap_reactants_and_products()
                return [reverse]
            case "both":
                forward, reverse = deepcopy(self), deepcopy(self)
                forward.name = forward.name + "[+]"
                reverse.name = reverse.name + "[-]"
                reverse.swap_reactants_and_products()
                return [forward, reverse]

    def swap_reactants_and_products(self):
        self.products, self.reactants = self.reactants, self.products

    def uniform_logic_notation(self, input: str) -> str:
        input = re.sub("\\|", ",", input)
        input = re.sub("\\*", "&", input)
        input = re.sub("\\s*\\(\\s*", "(", input)
        input = re.sub("\\s*\\)\\s*", ")", input)
        input = re.sub("\\s*&\\s*", "&", input)
        input = re.sub("\\s*,\\s*", ",", input)
        input = re.sub("\\s+", ",", input)
        return input



from __future__ import annotations
from tqdm import tqdm
import random
import copy
from algorithm.MaxRAFAlgorithm import MaxRAFAlgorithm
from algorithm.AlgorithmBase import AlgorithmBase
from model.ReactionSystem import ReactionSystem
from time import time

import sys
import os
sys.path.insert(0, os.path.abspath(
    os.path.join(os.path.dirname(__file__), '.')))


class MinIRAFHeuristic(AlgorithmBase):

    NAME: str = "iRAF"
    NUMBER_OF_RANDOM_INSERTION_ORDERS: int

    @property
    def number_of_random_insertion_orders(self):
        return self.NUMBER_OF_RANDOM_INSERTION_ORDERS

    @number_of_random_insertion_orders.setter
    def number_of_random_insertion_orders(self, value: int):
        self.NUMBER_OF_RANDOM_INSERTION_ORDERS = value

    @property
    def name(self):
        return self.NAME

    @name.setter
    def name(self, value: str):
        self.NAME = value

    @property
    def description(self):
        return "searches for irreducible RAFs in a heuristic fashion"

    def apply(self, input: ReactionSystem) -> ReactionSystem | None:
        """returns one of the smallest reaction systems possible produced by RAF

        Args:
            input (ReactionSystem): reaction system to be computed

        Returns:
            list[ReactionSystem]: reaction systems which are minimal
        """
        list = self.apply_all_smallest(input)
        if list:
            return list[0]
        else:
            return ReactionSystem()

    def apply_all_smallest(self, input: ReactionSystem) -> list[ReactionSystem]:
        """returns a list of the smallest reaction systems 

        Args:
            input (ReactionSystem): reaction system to be computed

        Returns:
            list[ReactionSystem]: reaction systems which are minimal
        """
        
        max_raf = MaxRAFAlgorithm().apply(input)
        
        if self.number_of_random_insertion_orders == None:
            self.number_of_random_insertion_orders = 10
        seeds = [
            i*123 for i in range(0, self.number_of_random_insertion_orders)]

        best: list[ReactionSystem] = []
        best_size = max_raf.size

        for seed in tqdm(seeds, desc="MinIRafHeuristic seeds: "):
            
            start_time = time()
            work_system = ReactionSystem(self.name)
            work_system.reactions = max_raf.reactions.copy()
            work_system.foods.update(max_raf.foods)
            ordering = max_raf.reactions.copy()
            random.Random(seed).shuffle(ordering)
            
            for reaction in ordering:
                try: work_system.reactions.remove(reaction)
                except: pass
                next = MaxRAFAlgorithm().apply(work_system)
                next.name = self.name
                tqdm.write("reactions run: " + str(time() -start_time))
                start_time = time()
                if next.size > 0 and next.size <= work_system.size:
                    work_system = next
                    if next.size < best_size:
                        best.clear()
                        best_size = next.size
                    if (next.size == best_size 
                        and not any([set(next.reactions) == set(a.reactions) 
                                     for a in best])):
                        best.append(next)
                    if best_size == 1:
                        break
                else:
                    work_system.reactions.append(reaction)
                
        if not best:
            result = copy.copy(max_raf)
            result.name = self.name
            best.append(result)

        return best



from __future__ import annotations
from enum import StrEnum


class ArrowNotation(StrEnum):

    __order__ = "USES_EQUALS USES_MINUS"
    
    USES_EQUALS = "USES_EQUALS"
    USES_MINUS = "USES_MINUS"

    def __init__(self, label: str = "=>") -> None:
        super().__init__()
        self.label: str = label

    def value_of_label(label: str) -> ArrowNotation | None:
        """returns appropriate enum-value given an arrow (label)

        Args:
            label (str): an arrow, i.e. '->' or '=>'

        Returns:
            ArrowNotation | None: USES_EQUALS or USES_MINUS
        """        
        if label in ["<=>", "<=", "=>"]:
            return ArrowNotation.USES_EQUALS
        elif label in ["<->", "<-", "->"]:
            return ArrowNotation.USES_MINUS
        else:
            return None



from __future__ import annotations
from model.ReactionSystem import ReactionSystem
'''
  AlgorithmBase.java Copyright (C) 2022 Daniel H. Huson
 
  (Some files contain contributions from other authors, who are then mentioned separately.)
 
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''
import sys
import os
sys.path.insert(0, os.path.abspath(
    os.path.join(os.path.dirname(__file__), '.')))


class AlgorithmBase:
    '''
    computes a new reaction system.
    super class of all algorithms
    '''

    NAME: str = ""

    @property
    def name(self):
        pass

    @property
    def description(self):
        '''
        get the name of the reaction system computed by this algorithm

        returns name
        '''
        pass

    def apply(input: ReactionSystem) -> ReactionSystem:
        '''
        run the algorithm

        return output
        '''

        pass

    def load_algorithms():
        pass

    def list_all_algorithms() -> set[str]:
        '''
        list all known algorithms

        return names of all known algorithms
        '''
        list = []
        path = os.path.dirname(os.path.abspath(__file__))

        for py in [f[:-3] for f in os.listdir(path) if f.endswith('.py') and f != '__init__.py' and f != "AlgorithmBase.py"]:
            parent_name = '.'.join(__name__.split('.')[:-1])
            mod = __import__('.'.join([parent_name, py]), fromlist=[py])
            classes = [getattr(mod, x) for x in dir(mod) if isinstance(getattr(mod, x), type)]
            for cls in classes:
                setattr(sys.modules[__name__], cls.__name__, cls)
        for algorithm in AlgorithmBase.__subclasses__():
            print(algorithm.NAME)
            list.append(algorithm.NAME)
        print(list)
        return set(list)

    def get_algorithm_by_name(name: str) -> AlgorithmBase | None:
        """gets a subclass of Algorithmbase by name

        Args:
            name (str): name of the searched algorithm

        Returns:
            AlgorithmBase | None: referenced algorithm
        """
        for algorithm in AlgorithmBase.__subclasses__():
            if name.casefold() == algorithm.NAME.casefold():
                return algorithm

        return None



from __future__ import annotations
from algorithm.AlgorithmBase import AlgorithmBase
from model.ReactionSystem import ReactionSystem
import Utilities as Utilities
from tqdm import tqdm
import sys
import os
sys.path.insert(0, os.path.abspath(
    os.path.join(os.path.dirname(__file__), '.')))


class MaxRAFAlgorithm (AlgorithmBase):

    NAME: str = "Max RAF"

    @property
    def name(self):
        return self.NAME

    @name.setter
    def name(self, value: str):
        self.NAME = value

    @property
    def description(self):
        return "computes the maximal RAF"

    def apply(self, input: ReactionSystem) -> ReactionSystem:
        """Computes the maximal RAF

        Args:
            input (ReactionSystem): system to be computed

        Returns:
            ReactionSystem: maximal RAF
        """       
        result = ReactionSystem(self.NAME)

        input_reactions = input.reactions.copy()
        input_food = input.foods.copy()

        if len(input_reactions) > 0:
            reactions = []
            molecules = []

            reactions.append(input_reactions)
            molecules.append(input_food)

            i = 0
            molecules.insert(
                i+1, Utilities.compute_closure(input_food, reactions[i]))
            reactions.insert(
                i+1, Utilities.filter_reactions(molecules[i+1], reactions[i]))

            while len(reactions[i+1]) < len(reactions[i]):
                i += 1
                molecules.insert(
                    i+1, Utilities.compute_closure(input_food, reactions[i]))
                reactions.insert(
                    i+1, Utilities.filter_reactions(molecules[i+1], reactions[i]))

            if len(reactions[i]) > 0:
                result.reactions = reactions[i]
                result.foods = result.compute_mentioned_foods(input.foods)

        return result



import os
import sys
sys.path.insert(0, os.path.abspath(
    os.path.join(os.path.dirname(__file__), '..')))
from model.ReactionSystem import ReactionSystem, MoleculeType, Reaction
from settings.ReactionNotation import ReactionNotation
from settings.ArrowNotation import ArrowNotation
from fileIO.IOManager import redirect_to_writer, OUTPUT_FILE_FORMATS, INPUT_FILE_FORMATS, ModelIO
from algorithm.MinIRAFHeuristic import MinIRAFHeuristic
from algorithm.MinRAFGeneratingElement import MinRAFGeneratingElement
from algorithm.AlgorithmBase import AlgorithmBase
import Utilities as Utilities
from itertools import combinations_with_replacement
import shutil
import random
import bisect
from tqdm import tqdm

from time import time
from os.path import isfile, join, isdir


ALL_ALGORITHMS = AlgorithmBase.list_all_algorithms()
ALL_ALGORITHMS.update([algo.lower()
                      for algo in AlgorithmBase.list_all_algorithms()])


def parse_input_file_to_rs(file_name: str) -> ReactionSystem:
    """Builds a ReactionSystem from a .crs-file

    Args:
        file_name (str): absolute path to .crs-file

    Raises:
        IOError: Wrong file Format
        IOError: Can't read file
        IOError: Can't detect format

    Returns:
        ReactionSystem: ReactionSystem equivalent to input_file
    """    
    res_lines = []
    notation = [None, None]
    if os.path.splitext(file_name)[1] not in INPUT_FILE_FORMATS:
        raise IOError("The file format " + os.path.splitext(file_name)[1] 
                      + " is not supported by this function.\n"
                      + "Please try a file with format in: " +
                      str(INPUT_FILE_FORMATS))
    with open(file_name, "r") as f:
        res_lines = f.readlines()
        if not res_lines:
            raise IOError("Can't read file: " + file_name)
        notation = ReactionNotation.detect_notation(res_lines)
        if notation[0] == None:
            raise IOError(
                "Can't detect 'full', 'sparse' or 'tabbed' file format.")
        res_reactiosystem = ReactionSystem()
        leading_comments = ModelIO.read(
            res_reactiosystem, file_name, notation[0])

        twr_str = "(" + str(res_reactiosystem.number_of_two_way_reactions) + \
            " two-way)" if res_reactiosystem.number_of_two_way_reactions > 0 else ""
        output_str = ("Read " + str(res_reactiosystem.size) + " reactions " +
                      twr_str + " and " + str(res_reactiosystem.food_size) +
                      " food items from file: " + file_name)
        if not leading_comments == "":
            output_str += "\n Comments:\n" + str(leading_comments)
        res_reactiosystem.update_inhibitors_present()
        if res_reactiosystem.inhibitors_present:
            output_str += "Input catalytic reaction system contains inhibitions. These are ignored in the computation of maxCAF, maxRAF and maxPseudoRAF"

        tqdm.write(output_str)

    return res_reactiosystem


def export_rs_to_output_file(output_systems:list[ReactionSystem],
                       output_path:str ="stdout",
                       output_format:str|None = None,
                       zipped:bool = False, 
                       reaction_notation:str = ReactionNotation.FULL,
                       arrow_notation:str = ArrowNotation.USES_EQUALS,
                       algorithm:AlgorithmBase|None = None):
    """export Reactionsystems as output_format-files

    Args:
        output_systems (list[ReactionSystem]): reaction systems to be written as a file/files
        output_path (str, optional): path to generated file. Default stdout prints in terminal. Defaults to "stdout".
        output_format (str | None, optional): If given this determines the file format of the output. Defaults to None.
        zipped (bool, optional): should the resulting files be zipped to an archive. Defaults to False.
        reaction_notation (str, optional): For '.crs' files. Determines reaction notation format. Defaults to ReactionNotation.FULL.
        arrow_notation (str, optional): For '.crs' files. Determines arrow format. Defaults to ArrowNotation.USES_EQUALS.
        algorithm (AlgorithmBase | None, optional): only relevant for message generation. Defaults to None.
    """
    redirect_to_writer(output_systems,
                       output_path,
                       output_format,
                       zipped, 
                       reaction_notation,
                       arrow_notation,
                       algorithm)


def apply_algorithm_to_rs(input_system: ReactionSystem,
                    algorithm: AlgorithmBase,
                    heuristic_runs: int = 10,
                    target_molecule:list[str] = []) -> ReactionSystem|None:
    """Applies the given algorithm to the input_system and returns the result.

    Args:
        input_system (ReactionSystem): system to find autocatalytic system in
        algorithm (AlgorithmBase): algorithm of autocatalytic system to find
        heuristic_runs (int, optional): number of seeds to check if algorithm is heuristic. Defaults to 10.
        target_molecule (list[str], optional): names of molecules to search for if algorithm has target. Defaults to [].

    Returns:
        ReactioSystem: autocatalytic ReactionSystem of given algorithm
    """    
    if isinstance(algorithm, MinIRAFHeuristic):
        algorithm.number_of_random_insertion_orders = heuristic_runs
        output_system = algorithm.apply(input_system)
    elif isinstance(algorithm, MinRAFGeneratingElement):
        algorithm.targets = MoleculeType().values_of(target_molecule)
        algorithm.number_of_random_insertion_orders = heuristic_runs
        output_system = algorithm.apply(input_system)
    else:
        output_system = algorithm.apply(input_system)

    return output_system


def apply_algorithm_to_file(algorithm: str | AlgorithmBase,
                  input_file: str,
                  output_path: str = "stdout",
                  zipped: bool = False,
                  output_format: str = ".crs",
                  reaction_notation: str | ReactionNotation = ReactionNotation.FULL,
                  arrow_notation: str | ArrowNotation = ArrowNotation.USES_EQUALS,
                  heuristic_runs: str | int = 10,
                  target_molecule:list[str] = [],
                  overwrite_ok: bool = False) -> ReactionSystem|None:
    """Applies the given algorithm to the input system, returns the result and writes it as a file at output_path.

    Args:
        algorithm (str | AlgorithmBase): (name of) algorithm of autocatalytic system to find
        input_file (str): path to .crs file to find autocatalytic system in
        output_path (str, optional): path to write resulting system to. Defaults to "stdout".
        zipped (bool, optional): Should the resulting file be zipped. Defaults to False.
        output_format (str, optional): file format of the output path. Overwrites output_file ending. Defaults to ".crs".
        reaction_notation (str | ReactionNotation, optional): Reaction format from FULL, SPARSE, TABBED. Defaults to ReactionNotation.FULL.
        arrow_notation (str | ArrowNotation, optional): Arrow format from USES_EQUALS, USES_MINUS. Defaults to ArrowNotation.USES_EQUALS.
        heuristic_runs (str | int, optional): number of seeds to check if algorithm is heuristic. Defaults to 10.
        target_molecule (list[str], optional): names of molecules to search for if algorithm has target. Defaults to [].
        overwrite_ok (bool, optional): Is the function allowed to write over existing files. Defaults to False.

    Returns:
        ReactionSystem|None: autocatalytic ReactionSystem of given algorithm
    """    
    start_time = time()
    input_system = parse_input_file_to_rs(input_file)

    if not overwrite_ok and os.path.exists(output_path):
        tqdm.write("The given output file " + output_path + " already exists.\n"
                   + "If you want to overwrite it please run this function with "
                   + "the 'overwrite_ok' attribute set to True")
        return

    if isinstance(algorithm, str):
        algorithm: AlgorithmBase = AlgorithmBase.get_algorithm_by_name(
            algorithm)
    
    if isinstance(algorithm, type): algorithm = algorithm()
    
    if isinstance(heuristic_runs, str):
        heuristic_runs = int(heuristic_runs)

    output_systems = apply_algorithm_to_rs(input_system, algorithm, heuristic_runs, target_molecule)

    redirect_to_writer([output_systems], output_path,
                       output_format, zipped,
                       reaction_notation, arrow_notation,
                       algorithm)
    
    tqdm.write("Total time spent: " + str(time() - start_time))
    return output_systems


def apply_algorithm_to_directory(algorithm: str | AlgorithmBase,
                       input_directory: str,
                       output_path: str = "stdout",
                       zipped: bool = False,
                       output_format: str = ".crs",
                       reaction_notation: str | ReactionNotation = ReactionNotation.FULL,
                       arrow_notation: str | ArrowNotation = ArrowNotation.USES_EQUALS,
                       heuristic_runs: str | int = 10,
                       target_molecule:list[str] = [],
                       overwrite_ok: bool = False,
                       time_dict:dict = {}) -> list[ReactionSystem]|None:
    """Applies the given algorithm to the input systems, returns the result and writes it as a file at output_path.

    Args:
        algorithm (str | AlgorithmBase): (name of) algorithm of autocatalytic system to find
        input_directory (str): path to directory to find autocatalytic systems in. Consumes all .crs files in the directory.
        output_path (str, optional): path to write resulting systems to. Defaults to "stdout".
        zipped (bool, optional): Should the resulting file be zipped. Defaults to False.
        output_format (str, optional): file format of the output path. Overwrites output_file ending. Defaults to ".crs".
        reaction_notation (str | ReactionNotation, optional): Reaction format from FULL, SPARSE, TABBED. Defaults to ReactionNotation.FULL.
        arrow_notation (str | ArrowNotation, optional): Arrow format from USES_EQUALS, USES_MINUS. Defaults to ArrowNotation.USES_EQUALS.
        heuristic_runs (str | int, optional): number of seeds to check if algorithm is heuristic. Defaults to 10.
        target_molecule (list[str], optional): names of molecules to search for if algorithm has target. Defaults to [].
        overwrite_ok (bool, optional): Is the function allowed to write over existing files. Defaults to False.

    Returns:
        list[ReactionSystem]|None: A list of all ReactionSystems created this way
    """    
    
    st = time()
    if os.path.splitext(output_path)[1]:
        output_path = os.path.splitext(output_path)[0]
    
    if output_format not in OUTPUT_FILE_FORMATS or not output_format:
        tqdm.write("please choose a valid file format from:\n" +
                   str(OUTPUT_FILE_FORMATS) + "\n"+
                   "File format cannot be inferred here.")
        return
    
    dir_files = [join(input_directory, f) for f in os.listdir(input_directory)
                 if isfile(join(input_directory, f))]

    file_zipped = False
    output_systems=[]
    for file in tqdm(dir_files, desc="Files to " + output_path):
        if os.path.splitext(file)[1] in INPUT_FILE_FORMATS:
            if isdir(output_path) or output_path == 'stdout':
                output_file = os.path.join(output_path, os.path.split(file)[1])
                output_systems.append(apply_algorithm_to_file(algorithm, file, output_file, file_zipped, output_format,
                              reaction_notation, arrow_notation, heuristic_runs, target_molecule, overwrite_ok))
            else:
                os.makedirs(output_path, exist_ok=True)
                output_file = os.path.join(output_path, os.path.split(file)[1])
                output_systems.append(apply_algorithm_to_file(algorithm, file, output_file, file_zipped, output_format,
                              reaction_notation, arrow_notation, heuristic_runs, target_molecule, overwrite_ok))
                
    if output_path != 'stdout' and zipped:
        shutil.make_archive(output_path, 'zip', output_path)
        if os.path.isdir(output_path):
            shutil.rmtree(output_path)
    et = time()
    time_dict.update({output_path:et-st})
    return output_systems


def generate_reaction_system_files(alphabet_sizes: list[int],
                                   food_max_lengths: list[int],
                                   polymer_max_lengths: list[int],
                                   means: list[float],
                                   number_of_replicates: int,
                                   output_directory: str = 'stdout',
                                   file_template: str = "polymer_model_a#a_k#k_n#n_m#m_r#r.crs",
                                   zipped: bool = False,
                                   output_format: str = '.crs',
                                   reaction_notation: ReactionNotation = ReactionNotation.FULL,
                                   arrow_notation: ArrowNotation = ArrowNotation.USES_EQUALS,
                                   overwrite_ok: bool = False) -> None:
    """Generates .crs files of random polymer systems given input constraints.

    Args:
        alphabet_sizes (list[int]): number of unique minimal elements. only supports < 11
        food_max_lengths (list[int]): maximum number of elements a food molecule may consist of
        polymer_max_lengths (list[int]): maximum number of elements any polymer molecule may consist of
        means (list[float]): mean number of reactions a polymer catalyzes
        number_of_replicates (int): seed for replication
        output_directory (str, optional): directory to write output systems into. Defaults to 'stdout'.
        file_template (str, optional): template for single files in dorectories. Defaults to "polymer_model_a#a_k#k_n#n_m#m_r#r.crs".
        zipped (bool, optional): should the resulting directory be zipped. Defaults to False.
        output_format (str, optional): file format of the output path. Overwrites file_template ending. Defaults to '.crs'.
        reaction_notation (ReactionNotation, optional): Reaction format from FULL, SPARSE, TABBED. Defaults to ReactionNotation.FULL.
        arrow_notation (ArrowNotation, optional): Arrow format from USES_EQUALS, USES_MINUS. Defaults to ArrowNotation.USES_EQUALS.
        overwrite_ok (bool, optional): Is the function allowed to write over existing files. Defaults to False.
    
    In the file template replacement works as follows:
        - "#a": alphabet_size
        - "#k": food_max_length
        - "#n": polymer_max_length
        - "#m": mean
        - "#r": number_of_replicates
    """    

    tqdm.write("writing files to: " + output_directory)
    
    files_count = 0
    total_files_expected = len(alphabet_sizes) * len(food_max_lengths) * len(
        polymer_max_lengths) * len(means) * len(number_of_replicates)
    with tqdm(total=total_files_expected, desc="Files: ") as pbar:
        for a in alphabet_sizes:
            for k in food_max_lengths:
                for n in polymer_max_lengths:
                    for m in means:
                        for r in number_of_replicates:

                            parameters = {"a": a, "k": k,
                                          "n": n, "m": m, "r": r}

                            if output_directory == "stdout":
                                filename = "stdout"
                            else:
                                file = Utilities.replace_parameters(
                                    file_template, parameters)
                                filename = os.path.join(output_directory, file)

                            if os.path.exists(filename) and not overwrite_ok:
                                tqdm.write("The given output file " + filename + " already exists.\n"
                                           + "If you want to overwrite it please run this function with "
                                           + "the 'overwrite_ok' attribute set to True")
                                continue

                            if all([x != None for x in parameters.values()]):
                                res_reaction_system = generate_reaction_system(
                                    parameters)
                            else:
                                tqdm.write(
                                    "Not all neccessary paramters were given.\n")
                                for key, value in zip(parameters.keys(), parameters.values()):
                                    if not value:
                                        tqdm.write("- missing: " + key)
                                return

                            redirect_to_writer([res_reaction_system], filename,
                                               output_format, zipped,
                                               arrow_notation, reaction_notation)

                            pbar.update(1)
                            files_count += 1
    tqdm.write("Number of files created: " + str(files_count))


def generate_reaction_system(parameters: dict[str:int, str:int, str:int, str:float, str:int]) -> ReactionSystem:
    """Generates a single random polymer reaction system given parameters

    Args:
        parameters (dict[str:int|float]): parameters of single random polymer system

    Returns:
        ReactionSystem: resulting reaction system.
        
    Parameters must include:
        - "a": alphabet_size (int)
        - "k": food_max_length (int)
        - "n": polymer_max_length (int)
        - "m": mean (float)
        - "r": number_of_replicates (int)
    """
    res_reaction_system = ReactionSystem(Utilities.replace_parameters(
        "# Polymer model a=#a k=#k n=#n m=#m r=#r", parameters))
    basic_elements = ["a" + str(i) for i in range(int(parameters["a"]))]
    food_names_lists = []
    for k in tqdm(range(1, int(parameters["k"]) + 1),
                  desc=res_reaction_system.name + " Foods Generation: "):
        food_names_lists.extend(
            list(combinations_with_replacement(basic_elements, k)))
    food_names: list[str] = ["".join(fl) for fl in food_names_lists]
    res_reaction_system.foods = MoleculeType().values_of(food_names)

    polymer_names_lists = []
    for n in tqdm(range(1, int(parameters["n"]) + 1),
                  desc=res_reaction_system.name + " Polymer Generation: "):
        polymer_names_lists.extend(
            list(combinations_with_replacement(basic_elements, n)))
    polymers: list[str] = ["".join(pl) for pl in polymer_names_lists]
    estimate_n_reactions = 0
    for polymer in polymers:
        estimate_n_reactions += len(polymer) / 2
    digits_estimate = len(str(int(estimate_n_reactions)))

    reactions = []
    count = 0
    with tqdm(desc=res_reaction_system.name + " Reaction Generation ",
              total=estimate_n_reactions) as r_pbar:
        for polymer in polymers:
            for i in range(0, len(polymer), 2):
                r_pbar.update(1)
                count += 1
                reaction = Reaction("r" + str(count).zfill(digits_estimate))
                if i == 0:
                    continue
                prefix = polymer[0:i]
                suffix = polymer[i:]
                reaction.reactants = MoleculeType().values_of([prefix, suffix])
                reaction.products = MoleculeType().values_of([polymer])
                reaction.direction = Reaction().DIRECTION["both"]
                reactions.append(reaction)

    reaction_length = len(reactions)
    p = parameters["m"]/reaction_length
    if p > 1:
        tqdm.write(
            "This mean: "
            + parameters["m"]
            + "\nis impossible to achieve since there are fewer possible reactions than the mean.")
        p = 1.0
    elif p < 0:
        tqdm.write(
            "This mean: "
            + parameters["m"]
            + "\nis impossible to achieve since it is negative.")
    dist = Utilities.binom_dist(
        reaction_length,  parameters["m"]/reaction_length)
    random.seed = parameters["r"]
    for polymer in polymers:
        successes = bisect.bisect_right(
            dist, random.uniform(0, 0.999999999999))
        i = 0
        while i < successes:
            reaction: Reaction = reactions[random.randint(
                0, reaction_length - 1)]
            if polymer in reaction.catalysts:
                continue
            if reaction.catalysts.strip() == "":
                reaction.catalysts = polymer
            else:
                reaction.catalysts += "," + polymer
            i += 1

    for reaction in reactions:
        if reaction.catalysts.strip() != "":
            res_reaction_system.reactions.append(reaction)

    return res_reaction_system



import os
import sys
sys.path.insert(0, os.path.abspath(
    os.path.join(os.path.dirname(__file__), '..')))

from fileIO.IOManager import OUTPUT_FILE_FORMATS, TRUTH_STRINGS
from algorithm.MinIRAFHeuristic import MinIRAFHeuristic
from algorithm.MinRAFGeneratingElement import MinRAFGeneratingElement
from tqdm import tqdm
import main.CatPyNet as cpn
import argparse
import importlib.metadata


#__version__ = importlib.metadata.version('catPyNet')

def main():

    parser = argparse.ArgumentParser(
        description="Performs Max RAF and other computations")
    
#    parser.add_argument("--version", action="version", version='%(prog)s ' + __version__)
    parser.add_argument("-c", metavar="compute", required=True,
                        help="The computation to perform from: " + str(cpn.ALL_ALGORITHMS), 
                        choices=cpn.ALL_ALGORITHMS)
    parser.add_argument("-i", metavar="input",
                        help="Input file (stdin ok)", default="stdin")
    parser.add_argument("-o", metavar='output_file',
                        help="Output file (stdout ok)", default="stdout")
    parser.add_argument("-z", metavar='output zipped',
                        help="Should the output be a zipped directory. (True or False)", 
                        choices=TRUTH_STRINGS, 
                        default="False")
    parser.add_argument("-of", metavar="output format",
                        help="file format to be written. e.g. '.crs'", 
                        choices=OUTPUT_FILE_FORMATS, 
                        default=None)
    parser.add_argument("-rn", metavar="reaction notation",
                        help="Output reaction notation", default="FULL")
    parser.add_argument("-an", metavar="arrow notation",
                        help="Output arrow notation", default="USES_EQUALS")
    parser.add_argument("-r", metavar="runs", help="Number of randomized runs for " +
                        MinIRAFHeuristic().name + " heuristic")
    parser.add_argument("-t", metavar="target molecule", help="Target Molecule for " +
                        MinRAFGeneratingElement().name, default="")
    parser.add_argument("-ow", metavar="overwrite ok", help="Sets if the program is allowed to " +
                        "write over files", choices=TRUTH_STRINGS, default="False")

    arguments = vars(parser.parse_args())
    zipped = True if arguments['z'].casefold() in ['True'.casefold(), "1"] else False
    overwrite_ok = True if arguments['ow'].casefold() in ['True'.casefold(), "1"] else False
    if arguments["i"] == 'stdin': 
        input_file = input("Please enter the file path you want to read from:")
    else:
        input_file = arguments["i"]
    if "," in arguments["t"]:
        target_molecules = arguments["t"].split(",")
    else:
        target_molecules = [arguments["t"]]
    
    if os.path.isdir(input_file):
        cpn.apply_algorithm_to_directory(arguments['c'], input_file, arguments['o'],zipped, arguments['of'],
                        arguments['rn'], arguments['an'], arguments['r'], target_molecules, 
                        overwrite_ok, time_dict={})
    elif os.path.isfile(input_file):
        cpn.apply_algorithm_to_file(arguments['c'], input_file, arguments['o'],zipped, arguments['of'],
                        arguments['rn'], arguments['an'], arguments['r'], target_molecules, 
                        overwrite_ok)
    else:
        raise IOError("Please select an existing file or directory as input.\n"
                      + "You tried to select: " + input_file)

if __name__ == "__main__":
    main()



from __future__ import annotations
from algorithm.MaxRAFAlgorithm import MaxRAFAlgorithm
from algorithm.Importance import Importance
from algorithm.AlgorithmBase import AlgorithmBase
from model.ReactionSystem import ReactionSystem
from tqdm import tqdm

import sys
import os
sys.path.insert(0, os.path.abspath(
    os.path.join(os.path.dirname(__file__), '.')))


class CoreRAFAlgorithm (AlgorithmBase):
    '''
    Computes the unique, irreducible RAF if it exists.
    '''
    NAME: str = "Core RAF"

    @property
    def name(self) -> str:
        return self.NAME

    @property
    def description(self):
        return "computes the unique irreducible RAF, if it exists (Section 4.1 of [SXH20])"

    def apply(self, input: ReactionSystem) -> ReactionSystem:
        """Computes the unique, irreducible RAF if it exists.

        Args:
            input (ReactionSystem): reaction system to reduce

        Returns:
            ReactionSystem: core raf
        """
        max_raf = MaxRAFAlgorithm().apply(input)
        with tqdm(total=max_raf.size, desc="CoreRAFAlgorithm") as craf_pbar:

            important_reactions = ReactionSystem("Important")
            important_reactions.reactions = [p[0] for p in Importance.compute_reaction_importance(
                input, max_raf, MaxRAFAlgorithm, craf_pbar) if p[1] == 100.0]
            important_reactions.foods = important_reactions.compute_mentioned_foods(input.foods)

            core_raf = MaxRAFAlgorithm().apply(important_reactions)
            core_raf.name = "Core RAF"
            craf_pbar.update(1)
        return core_raf



# -*- coding: utf-8 -*-

import sys
import os
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
from fileIO.IOManager import OUTPUT_FILE_FORMATS, INPUT_FILE_FORMATS
from settings.ReactionNotation import ReactionNotation, ArrowNotation
import tools.CommandLineTool as clt
import main.CatPyNet as cpn
import fileIO.GraphIO as gio
from algorithm.AlgorithmBase import AlgorithmBase
from algorithm.MaxRAFAlgorithm import MaxRAFAlgorithm
from algorithm.MaxCAFAlgorithm import MaxCAFAlgorithm
from algorithm.MinIRAFHeuristic import MinIRAFHeuristic
from algorithm.CoreRAFAlgorithm import CoreRAFAlgorithm
from algorithm.MaxPseudoRAFAlgorithm import MaxPseudoRAFAlgorithm
from os import listdir
from os.path import isfile, join
from tqdm import tqdm

algos = AlgorithmBase.list_all_algorithms()
mypath = "G:\\Github\\BA-Jan\\test_files"
respath = "D:\\Users\\jrls2_000\\Documents\\UNI\\_BA\\test_data_results"
test_files = [f for f in listdir(mypath) if isfile(join(mypath, f)) 
              and os.path.splitext(f)[1] in INPUT_FILE_FORMATS]

real_res = {"Max RAF":{"example-0.crs":{"food":3,
                                        "reactions":["r1", "r2", "r3"]},
                       "example-1.crs":{"food":12,
                                        "reactions":["r1", "r2", "r3", "r'1", "r'2", "r'3"]},
                       "example-10.crs":{"food":2,
                                        "reactions":["r1", "r2", "r3"]},
                       "example-2.crs":{"food":5,
                                        "reactions":["r1", "r2", "r3", "r4", "r5"]},
                       "example-3.crs":{"food":1,
                                        "reactions":["r1", "r2", "r3","r'1", "r'2", "r'3",
                                                     "r#1", "r#2", "r#3", "r#'1", "r#'2", "r#'3",
                                                     "r_theta", "r_x"]},
                       "example-4.crs":{"food":1,
                                        "reactions":["r1", "r2", "r3", "r4", "r5", "r6",
                                                     "r7", "r8", "r9"]},
                       "example-5.crs":{"food":2,
                                        "reactions":["r11", "r12", "r13", "r14",
                                                     "r21", "r22", "r23", "r24",
                                                     "r31", "r32", "r33", "r34",
                                                     "r41", "r42", "r43", "r44"]},
                       "example-6.crs":{"food":1,
                                        "reactions":["r1", "r2", "r3", "r4", "r5", "r6","r7"]},
                       "example-7.crs":{"food":4,
                                        "reactions":["r1", "r2", "r3"]},
                       "example-8.crs":{"food":4,
                                        "reactions":["r01", "r02", "r03", "r04", "r05", "r06",
                                                     "r07", "r08", "r09", "r10",
                                                     "r11", "r12", "r13", "r14", "r15", "r16",
                                                     "r17"]},
                       "example-9.crs":{"food":5,
                                        "reactions":["r1", "r2", "r3", "r4"]},
                       "inhibitions-1.crs":{"food":3,
                                        "reactions":["r1"]}},
            "Max CAF":{"example-0.crs":{"food":0,
                                        "reactions":[]},
                       "example-1.crs":{"food":0,
                                        "reactions":[]},
                       "example-10.crs":{"food":2,
                                        "reactions":["r1", "r2", "r3"]},
                       "example-2.crs":{"food":4,
                                        "reactions":["r3", "r4", "r5"]},
                       "example-3.crs":{"food":0,
                                        "reactions":[]},
                       "example-4.crs":{"food":0,
                                        "reactions":[]},
                       "example-5.crs":{"food":0,
                                        "reactions":[]},
                       "example-6.crs":{"food":0,
                                        "reactions":[]},
                       "example-7.crs":{"food":4,
                                        "reactions":["r1", "r2", "r3"]},
                       "example-8.crs":{"food":4,
                                        "reactions":["r01", "r02", "r03", "r04", "r05", "r06",
                                                     "r07", "r08", "r09", "r10",
                                                     "r11", "r12", "r13", "r14", "r15", "r16",
                                                     "r17"]},
                       "example-9.crs":{"food":0,
                                        "reactions":[]},
                       "inhibitions-1.crs":{"food":3,
                                        "reactions":["r1"]}},
            "Max Pseudo RAF":{"example-0.crs":{"food":3,
                                        "reactions":["r1", "r2", "r3", "r4", "r5", "r6"]},
                       "example-1.crs":{"food":12,
                                        "reactions":["r1", "r2", "r3", "r'1", "r'2", "r'3"]},
                       "example-10.crs":{"food":2,
                                        "reactions":["r1", "r2", "r3"]},
                       "example-2.crs":{"food":5,
                                        "reactions":["r1", "r2", "r3", "r4", "r5"]},
                       "example-3.crs":{"food":1,
                                        "reactions":["r1", "r2", "r3","r'1", "r'2", "r'3",
                                                     "r#1", "r#2", "r#3", "r#'1", "r#'2", "r#'3",
                                                     "r_theta", "r_x"]},
                       "example-4.crs":{"food":1,
                                        "reactions":["r1", "r2", "r3", "r4", "r5", "r6",
                                                     "r7", "r8", "r9"]},
                       "example-5.crs":{"food":2,
                                        "reactions":["r11", "r12", "r13", "r14",
                                                     "r21", "r22", "r23", "r24",
                                                     "r31", "r32", "r33", "r34",
                                                     "r41", "r42", "r43", "r44"]},
                       "example-6.crs":{"food":1,
                                        "reactions":["r1", "r2", "r3", "r4", "r5", "r6","r7"]},
                       "example-7.crs":{"food":4,
                                        "reactions":["r1", "r2", "r3"]},
                       "example-8.crs":{"food":4,
                                        "reactions":["r01", "r02", "r03", "r04", "r05", "r06",
                                                     "r07", "r08", "r09", "r10",
                                                     "r11", "r12", "r13", "r14", "r15", "r16",
                                                     "r17"]},
                       "example-9.crs":{"food":6,
                                        "reactions":["r1", "r2", "r3", "r4", "r5", "r6",
                                                     "r7"]},
                       "inhibitions-1.crs":{"food":3,
                                        "reactions":["r1"]}},
            "iRAF":{"example-0.crs":{"food":3,
                                        "reactions":3},
                       "example-1.crs":{"food":12,
                                        "reactions":3},
                       "example-10.crs":{"food":2,
                                        "reactions":1},
                       "example-2.crs":{"food":5,
                                        "reactions":1},
                       "example-3.crs":{"food":1,
                                        "reactions":8},
                       "example-4.crs":{"food":1,
                                        "reactions":1},
                       "example-5.crs":{"food":2,
                                        "reactions":1},
                       "example-6.crs":{"food":1,
                                        "reactions":1},
                       "example-7.crs":{"food":4,
                                        "reactions":1},
                       "example-8.crs":{"food":4,
                                        "reactions":1},
                       "example-9.crs":{"food":6,
                                        "reactions":4},
                       "inhibitions-1.crs":{"food":3,
                                        "reactions":1},
                       "prokaryotic-network.crs":{"food":3,
                                        "reactions":1}}}

def run_everything():
    #algos.remove("iRAF")
    #algos = ["iRAF"]
    total_test_files = (len(algos) * 2 * (len(OUTPUT_FILE_FORMATS) - 1) * len(ReactionNotation) 
                        * len(ArrowNotation) * len(test_files))
    with tqdm(desc="Total Test Files:", total=total_test_files) as tot_f:
        for algo in algos:
            for zipped in [True, False]:
                for output_format in OUTPUT_FILE_FORMATS:
                    for reaction_notation in ReactionNotation:
                        for arrow_notation in ArrowNotation:
                            if not output_format: continue
                            output_addition = os.path.join(algo, 
                                                        str(zipped), 
                                                        output_format.removeprefix("."), 
                                                        reaction_notation.value,
                                                        arrow_notation.value)
                            
                            output_path = os.path.join(respath, output_addition)
                            
                            cpn.apply_algorithm_to_directory(algo,
                                                mypath,
                                                output_path,
                                                zipped,
                                                output_format,
                                                reaction_notation,
                                                arrow_notation,
                                                100,
                                                True)
                            
                            tot_f.update(len(test_files))
                            
def test_one_algo(algo:AlgorithmBase):
    total_test_files = (len(ReactionNotation) * len(ArrowNotation) * len(test_files))
    output_format = ".crs"
    time_dict = {}
    with tqdm(desc=algo.NAME + " Test Files:", total=total_test_files) as tot_f:
        for reaction_notation in ReactionNotation:
            for arrow_notation in ArrowNotation:
                if not output_format: continue
                output_addition = os.path.join(algo.NAME,
                                            output_format.removeprefix("."), 
                                            reaction_notation.value,
                                            arrow_notation.value)
                
                output_path = os.path.join(respath, output_addition)
                
                output_systems = cpn.apply_algorithm_to_directory(algo,
                                    mypath,
                                    output_path,
                                    output_format=output_format,
                                    zipped=False,
                                    reaction_notation=reaction_notation,
                                    arrow_notation=arrow_notation,
                                    overwrite_ok=True,
                                    heuristic_runs=100,
                                    time_dict=time_dict)
                
                tot_f.update(len(test_files))
    output_foods = {}
    output_reactions = {}
    food_truth = {}
    reaction_truth = {}
    filenames = [os.path.split(file)[1] for file in test_files]
    for i, rs in enumerate(output_systems):
        filename = filenames[i]
        output_foods.update({filename:rs.food_size})
        output_reactions.update({filename:[r.name for r in rs.reactions]})
    if algo.NAME in ["Max RAF", "Max CAF", "Max Pseudo RAF"]:
        for file in filenames:
            file_dict = real_res[algo.NAME][file]
            food_truth.update({file:file_dict["food"] == output_foods[file]})
            tqdm.write(file)
            tqdm.write("Food: " + str(file_dict["food"] == output_foods[file]))
            file_dict_reactions = file_dict["reactions"]
            file_dict_reactions.sort()
            output_reactions[file].sort()
            reaction_truth.update({file:file_dict_reactions == output_reactions[file]})
            tqdm.write("Reactions: " + str(file_dict_reactions == output_reactions[file]))
    if algo.NAME in ["iRAF"]:
        for file in filenames:
            file_dict = real_res[algo.NAME][file]
            food_truth.update({file:file_dict["food"] == output_foods[file]})
            tqdm.write(file)
            tqdm.write("Food: " + str(file_dict["food"] == output_foods[file]))
            file_dict_reactions = file_dict["reactions"]
            reaction_truth.update({file:file_dict_reactions == len(output_reactions[file])})
            tqdm.write("Reactions: " + str(file_dict_reactions == len(output_reactions[file])))
    
        tqdm.write("Food: ")
        for key, value in zip(food_truth.keys(), food_truth.values()):
            if not value: tqdm.write(key)
        
        tqdm.write("Reactions: ")
        for key, value in zip(reaction_truth.keys(), reaction_truth.values()):
            if not value: tqdm.write(key)
        
    
    tqdm.write(str(float(sum(time_dict.values())) / float(len(time_dict.values()))))
    return all(food_truth.values()) and all(reaction_truth.values())
    
    
        

if __name__ == "__main__":
    
    reaction_system = cpn.parse_input_file_to_rs(os.path.join(mypath, "inhibitions-1.crs"))
    gio.print_rs_as_graph(reaction_system)
    #tqdm.write(str(test_one_algo(MinIRAFHeuristic)))
    #run_everything()           
    
    
    
    """ for algo in algos:
        if algo=="iRAF":continue
        algo_respath = respath + "\\" + algo
        for j, file in enumerate(test_files):
            if j == 14: continue
            sys.argv.append("-c")
            sys.argv.append(algo)
            sys.argv.append("-z")
            sys.argv.append("False")
            sys.argv.append("-of")
            sys.argv.append(".crs")
            sys.argv.append("-i")
            sys.argv.append(mypath + "\\"+ file)
            sys.argv.append("-o")
            full_file = algo_respath + "\\" + "result-" + file
            os.makedirs(os.path.dirname(full_file), exist_ok=True)
            sys.argv.append(full_file)
            clt.main()
            for i in range(1, len(sys.argv)):
                del sys.argv[1]
            gc.collect() """
    
    print([[chr(v) for v in range(ord("a"), ord("a") + 26)]])



from __future__ import annotations
from model.MoleculeType import MoleculeType
from model.Reaction import Reaction
import sys
import os
sys.path.insert(0, os.path.abspath(
    os.path.join(os.path.dirname(__file__), '.')))


class ReactionSystem:
    '''
    A catalytic reaction system
    '''

    def __init__(self, name: str = "Reactions", **kwargs):
        """Create a reaction system.

        Args:
            name (str, optional): Name of the reaction system. Defaults to "Reactions".

        KwArgs:
            reactions (list[Reaction]): reactions in reaction system. Defaults to [].
            foods (list[MoleculeType]): food in reaction system. Defaults to [].
        """
        self.name: str = name
        self.reactions: list[Reaction] = [
        ] if "reactions" not in kwargs else kwargs["reactions"]
        self.foods: set[MoleculeType] = set(
        ) if "foods" not in kwargs else kwargs["foods"]
        self.inhibitors_present: bool = False
        self.size: int
        self.food_size: int
        buffer = 0
        for reaction in self.reactions:
            if reaction.direction == "both":
                buffer += 1
        self.number_of_two_way_reactions: int = buffer

        self.update_inhibitors_present()

    @property
    def reactions(self):
        return self._reactions

    @reactions.setter
    def reactions(self, value: list[Reaction]):
        self._reactions = value
        buffer = 0
        for reaction in value:
            if reaction.direction == "both":
                buffer += 1
        self._number_of_two_way_reactions = buffer

    @property
    def size(self):
        return len(self.reactions)

    @property
    def food_size(self):
        return len(self.foods)

    def __copy__(self) -> ReactionSystem:
        res = ReactionSystem(self.name)
        res.foods = self.foods
        res.reactions = self.reactions
        return res

    def clear(self) -> None:
        self.reactions.clear()
        self.foods.clear()

    def get_header_line(self) -> str:
        res = [self.name, " has ", str(self.size)]
        if (self.get_number_of_one_way_reactions() == 0
                and self.number_of_two_way_reactions > 0):
            res.append(" two-way reactions")
        elif (self.get_number_of_one_way_reactions() > 0
              and self.number_of_two_way_reactions == 0):
            res.append(" one-way reactions")
        elif (self.get_number_of_one_way_reactions() > 0
              and self.number_of_two_way_reactions > 0):
            res.append(" reactions (")
            res.append(str(self.number_of_two_way_reactions))
            res.append(" two-way and ")
            res.append(str(self.get_number_of_one_way_reactions()))
            res.append(" one-way)")
        else:
            res.append(" reactions")
        res.append(" on ")
        res.append(str(len(self.foods)))
        res.append(" food items")
        return "".join(res)

    def update_inhibitors_present(self) -> None:
        """Sets inhibitors_present to True if any inhibitors are present.

        Checks all reactions and sets inhibitors_present to true if any
        inhibitor for any reaction is present.
        Otherwise sets inhibitors_present to False
        """
        for reaction in self.reactions:
            if len(reaction.inhibitions) > 0:
                self.inhibitors_present = True
                return
        self.inhibitors_present = False

    def get_mentioned_molecules(self) -> set[MoleculeType]:
        """gets all molecules mentioned in self

        Returns:
            set[MoleculeType]: reactants, products, inhibitions, catalysts
        """
        molecule_types = self.foods
        for reaction in self.reactions:
            molecule_types.update(reaction.reactants)
            molecule_types.update(reaction.products)
            molecule_types.update(reaction.inhibitions)
            catalysts = MoleculeType().values_of(reaction.catalysts.replace(",", "\t")
                                                 .replace("|", "\t").replace("*", "\t")
                                                 .replace("&", "\t").split("\t"))
            molecule_types.update(catalysts)
        return molecule_types

    def get_reaction_names(self) -> set[str]:
        names = []
        for reaction in self.reactions:
            names.append(reaction.name)
        return names

    def compute_mentioned_foods(self, foods: set[MoleculeType]) -> set[MoleculeType]:
        """Returns all food items that are mentioned in reactions

        Args:
            foods (set[MoleculeType]): foods to compare reactions against

        Returns:
            set[MoleculeType]: set of all MoleculeTypes mentioned in any reaction and foods
        """
        molecule_types = set()
        for reaction in self.reactions:
            molecule_types.update(reaction.reactants)
            molecule_types.update(reaction.inhibitions)
            molecule_types.update(reaction.products)
            molecule_types.update(reaction.get_catalyst_elements())
        return molecule_types.intersection(foods)

    def replace_named_reaction(self, name: str, reaction: Reaction) -> None:
        old_reaction = self.get_reaction_by_name(name)
        if old_reaction == None:
            raise TypeError("no such reaction: " + name)
        self.reactions.remove(old_reaction)
        self.reactions.append(reaction)

    def get_reaction_by_name(self, name: str) -> Reaction:
        for reaction in self.reactions:  # UNSCHÖN
            if reaction.name == name:
                return reaction

    def get_number_of_one_way_reactions(self) -> int:
        return self.size - self.number_of_two_way_reactions

    def __eq__(self, other: ReactionSystem) -> bool:
        if hash(self) == hash(other):
            return True
        if not (isinstance(other, ReactionSystem)):
            return False
        return (self.foods == other.foods
                and set(self.reactions) == set(other.reactions))

    def __lt__(self, other: Reaction) -> bool:
        return hash(self) < hash(other)

    def __hash__(self) -> int:
        return hash((tuple(self.foods), tuple(self.reactions)))



from __future__ import annotations

from copy import deepcopy


class MoleculeType:
    '''
    a molecule type
    '''

    def __deepcopy__(self, memo) -> MoleculeType:
        id_self = id(self)
        _copy = memo.get(id_self)
        if _copy == None:
            _copy = MoleculeType(deepcopy(self.name, memo),
                                 deepcopy(self.name2type, memo))
            memo[id_self] = _copy
        return _copy

    def __init__(self, name: str = None, name2type: dict = {}) -> None:
        self.name = name
        self.name2type = name2type

    def value_of(self, name: str) -> MoleculeType:
        if not (name in self.name2type):
            self.name2type[name] = MoleculeType(name)
        return self.name2type[name]

    def values_of(self, names: list[str]) -> set[MoleculeType]:
        return set([self.value_of(name) for name in names])

    def __eq__(self, other: MoleculeType) -> bool:
        if isinstance(other, MoleculeType):
            return (self.name == other.name)

    def __lt__(self, other: MoleculeType) -> bool:
        return hash(self) < hash(other)

    def __hash__(self) -> int:
        return hash(self.name)

    def __str__(self) -> str:
        return self.name



import sys
import os
sys.path.insert(0, os.path.abspath(
    os.path.join(os.path.dirname(__file__), '.')))

food_file = "D:\\Users\\jrls2_000\\Documents\\UNI\\_BA\\Data\\food.txt"
test_file = "D:\\Users\\jrls2_000\\Documents\\UNI\\_BA\\Data\\rsta20210244_si_002.txt"
target_file = os.path.join("G:\\Github\\BA-Jan\\test_results", "test_reactions.crs")

def main():
    foods = []
    with open(food_file, "r") as f_f:
        foods = f_f.readlines()
        for j, food in enumerate(foods):
            foods[j] = food.strip()
    
    with open(test_file, 'r') as f:
        lines = f.readlines()
        foodline = 0
        for i, line in enumerate(lines):
            if line.startswith("**"): lines[i] = "# " + line; continue
            if line == "\n": lines[i] = "# " + line ; continue
            if line.startswith('Reaction_ID'): lines[i] = "# " + line; foodline = i+1; continue
            splitline = line.split("\t")
            cat = "[" + splitline[2].strip() + "]"
            splitline = splitline[:2]
            splitline.append(cat)
            lines[i] = "\t".join(splitline) + "\n"

        lines.insert(foodline, "Food: " + ",".join(foods))
        with open(target_file, 'w') as target_f:
            target_f.writelines(lines)

if __name__ == '__main__':
    main()


from __future__ import annotations
from algorithm.AlgorithmBase import AlgorithmBase
from model.MoleculeType import MoleculeType
from model.Reaction import Reaction
from model.ReactionSystem import ReactionSystem
import time
from tqdm import tqdm

import copy
import sys
import os
sys.path.insert(0, os.path.abspath(
    os.path.join(os.path.dirname(__file__), '.')))


class Importance ():

    def get_description() -> str:
        return "computes the percent difference between model size and model size without given food item [HS23]"

    def compute_food_importance(input_system: ReactionSystem,
                                original_result: ReactionSystem,
                                algorithm: AlgorithmBase) -> list[tuple[MoleculeType, float]]:
        """Calculates importance of one molecule in food for reaction system size.

        Args:
            input_system (ReactionSystem): system to be analyzed
            original_result (ReactionSystem): original result of algorithm on input system
            algorithm (AlgorithmBase): algorithm to be used on input system

        Returns:
            list[tuple[MoleculeType, float]]: list of importances of food molecules from (0, 100].
            
        Food items are only listed if they have any importance. If they can be ignored with the same result they are not listed.
        """        
        result = []
        for food in input_system.foods:
            replicate_input = copy.copy(input_system)
            replicate_input.name = "Food importance"
            replicate_input.foods.remove(food)

            replicate_output = algorithm().apply(replicate_input)
            importance = 100.0 * \
                (original_result.size - replicate_output.size) / \
                float(original_result.size)
            if importance > 0:
                result.append((food, importance))
            result.sort(lambda x: x[1])
        return result

    def compute_reaction_importance(input_system: ReactionSystem,
                                    original_result: ReactionSystem,
                                    algorithm: AlgorithmBase,
                                    pbar: tqdm) -> list[tuple[Reaction, float]]:
        """Calculates importance of one reacton for reaction system size.

        Args:
            input_system (ReactionSystem): system to be analyzed
            original_result (ReactionSystem): original result of algorithm on input system
            algorithm (AlgorithmBase): algorithm to be used on input system

        Returns:
            list[tuple[MoleculeType, float]]: list of importances of reactions from (0, 100].
            
        Reaction items are only listed if they have any importance. If they can be ignored with the same result they are not listed.
        """          
        result = []
        if original_result.size == 1:
            result.append((original_result.reactions[0], 100.0))
        elif original_result.size > 1:
            size_to_compare_against = original_result.size - 1
            replicate_input = ReactionSystem("Reaction importance")
            replicate_input.foods = input_system.foods
            replicate_input.reactions = input_system.reactions.copy()
            for reaction in tqdm(input_system.reactions, desc="Calculating Reaction Importance: "):
                replicate_input.reactions.remove(reaction)

                replicate_output: ReactionSystem = algorithm().apply(replicate_input)
                if replicate_output.size < size_to_compare_against:
                    importance = (100.0 * 
                        (size_to_compare_against - replicate_output.size) / 
                        float(size_to_compare_against))
                    if importance > 0:
                        result.append((reaction, importance))
                replicate_input.reactions.append(reaction)
                pbar.update(1)
            result.sort(key=lambda x: x[1])
        return result

    def to_string_food_importance(food_imortance: list[tuple[MoleculeType, float]]) -> str:
        buffer = ""
        buffer += "Food importance: "
        first = True
        for pair in food_imortance:
            if first:
                first = False
            else:
                buffer += ", "
            buffer += pair[0].name + " " + str(pair[1])

        return buffer

    def to_string_reaction_importance(reaction_imortance: list[tuple[MoleculeType, float]]) -> str:
        buffer = ""
        buffer += "Reaction importance: "
        first = True
        for pair in reaction_imortance:
            if first:
                first = False
            else:
                buffer += ", "
            buffer += pair[0].name + " " + str(pair[1])

        return buffer



from model.ReactionSystem import ReactionSystem
from settings.ReactionNotation import ReactionNotation
from settings.ArrowNotation import ArrowNotation
from algorithm.AlgorithmBase import AlgorithmBase
from fileIO.GraphIO import write, SUPPORTED_GRAPH_OUTPUT_FILE_FORMATS
from fileIO.ModelIO import ModelIO, SUPPORTED_OUTPUT_FILE_FORMATS, SUPPORTED_INPUT_FILE_FORMATS
import shutil
from tqdm import tqdm

import os
import sys
sys.path.insert(0, os.path.abspath(
    os.path.join(os.path.dirname(__file__), '..')))


OUTPUT_FILE_FORMATS = SUPPORTED_OUTPUT_FILE_FORMATS
OUTPUT_FILE_FORMATS.update(SUPPORTED_GRAPH_OUTPUT_FILE_FORMATS)
OUTPUT_FILE_FORMATS.update(set([format.casefold() for format in OUTPUT_FILE_FORMATS]))
OUTPUT_FILE_FORMATS.update([None])

INPUT_FILE_FORMATS = SUPPORTED_INPUT_FILE_FORMATS
#INPUT_FILE_FORMATS.update(SUPPORTED_GRAPH_OUTPUT_FILE_FORMATS)
INPUT_FILE_FORMATS.update(set([format.casefold() for format in INPUT_FILE_FORMATS]))

TRUTH_STRINGS = ["True", "False", "1", "0"]

def redirect_to_writer(output_systems:list[ReactionSystem],
                       output_path:str ="stdout",
                       output_format:str|None = None,
                       zipped:bool = False, 
                       reaction_notation:str = ReactionNotation.FULL,
                       arrow_notation:str = ArrowNotation.USES_EQUALS,
                       algorithm:AlgorithmBase|None = None) -> None:
    """Processes the inputs and redirects data to appropriate file writer.

    Args:
        output_systems (list[ReactionSystem]): reaction systems to be written as a file/files
        output_path (str, optional): path to generated file. Default stdout prints in terminal. Defaults to "stdout".
        output_format (str | None, optional): If given this determines the file format of the output. Defaults to None.
        zipped (bool, optional): should the resulting files be zipped to an archive. Defaults to False.
        reaction_notation (str, optional): For '.crs' files. Determines reaction notation format. Defaults to ReactionNotation.FULL.
        arrow_notation (str, optional): For '.crs' files. Determines arrow format. Defaults to ArrowNotation.USES_EQUALS.
        algorithm (AlgorithmBase | None, optional): only relevant for message generation. Defaults to None.
    """
    if output_format and output_path != "stdout":
        output_path = os.path.splitext(output_path)[0] + output_format
    elif not output_format and output_path != "stdout":
        output_format = os.path.splitext(output_path)[1]

    if zipped:
        output = os.path.split(os.path.abspath(output_path))
        output_directory = os.path.join(output[0], output[1].split(".")[0])
        output_file = output[1]
        output_path = os.path.join(output_directory, output_file)
        os.makedirs(os.path.dirname(output_path), exist_ok=True)

    error_str = "The resulting reaction systems have no reactions.\n"
    if algorithm:
        error_str += "No " + algorithm.NAME
    if isinstance(output_systems, list):
        if all(output_systems):
            if not any([rs.reactions for rs in output_systems]):
                tqdm.write(error_str)
                return
        if not any(output_systems):
            tqdm.write(error_str)
            return
    else:
        output_systems = [output_systems]
        if not any([rs.reactions for rs in output_systems]):
            tqdm.write(error_str)
            return
        
    if output_format in SUPPORTED_GRAPH_OUTPUT_FILE_FORMATS:
        write(output_systems, output_path, output_format)
    elif output_format == ".crs":
        os.makedirs(os.path.dirname(output_path), exist_ok=True)
        with open(output_path, "w") as f:
            res_str = ""
            for output_system in output_systems:
                if not output_system.reactions:
                    res_str += ("The resulting reaction system has no reactions.\n"
                        + "No " + algorithm.NAME)
                else:
                    tqdm.write(output_system.get_header_line())
                    res_str += ModelIO().write(output_system,
                                            True,
                                            reaction_notation,
                                            arrow_notation)
                res_str += "\n"
            tqdm.write("Writing file: " + output_path)
            f.write(res_str)
    elif output_path == "stdout":
        res_str = ""
        for output_system in output_systems:
            if not output_system.reactions:
                res_str += ("The resulting reaction system has no reactions.\n"
                       + "No " + algorithm.NAME)
            else:
                tqdm.write(output_system.get_header_line())
                res_str += ModelIO().write(output_system,
                                        True,
                                        reaction_notation,
                                        arrow_notation)
            res_str += "\n"
        tqdm.write(res_str)
    else:
        tqdm.write("Given output file format was not recognized.\n" +
                   ".crs is assumed.")
        output_path = output_path.split(".")[0] + ".crs"
        os.makedirs(os.path.dirname(output_path), exist_ok=True)
        for output_system in output_systems:
            with open(output_path, "w") as f:
                res_str = ""
                if not output_system.reactions:
                        res_str += ("The resulting reaction system has no reactions.\n"
                            + "No " + algorithm.NAME)
                else:
                    tqdm.write(output_system.get_header_line())
                    res_str += ModelIO().write(output_system,
                                            True,
                                            reaction_notation,
                                            arrow_notation)
                res_str += "\n"
                tqdm.write("Writing file: " + output_path)
                f.write(res_str)

    if zipped:
        shutil.make_archive(output_directory, 'zip', output_directory)
        if os.path.isdir(output_directory):
            shutil.rmtree(output_directory)
    


import os
import sys
sys.path.insert(0, os.path.abspath(
    os.path.join(os.path.dirname(__file__), '..')))

from tqdm import tqdm
import main.CatPyNet as cpn
import argparse
from fileIO.IOManager import OUTPUT_FILE_FORMATS, TRUTH_STRINGS
import Utilities as Ut



def main():

    parser = argparse.ArgumentParser(description="Constructs Polymer Models")
    parser.add_argument("-a", metavar="alphabet size",
                        help="alphabet size (list (x,y,z,...) or range (x-z or x-z/step) ok)",
                        default="2")
    parser.add_argument("-k", metavar="food max length",
                        help="food molecule max length  (list or range ok)", default="2")
    parser.add_argument("-n", metavar='polymer max length',
                        help="polymer max length  (list or range ok)", default="4")
    parser.add_argument("-m", metavar="mean catalyzed",
                        help="mean number of catalyzed reactions per molecule  (list or range ok)",
                        default="2.0")
    parser.add_argument("-r", metavar="replicate",
                        help="The replicate number/seed (list or range ok)", default="1")
    parser.add_argument("-o", metavar="output", help="Output directory (or stdout)",
                        default="stdout")
    parser.add_argument("-f", metavar="file name template",
                        help="file name template (use #a,#k,#n,#m,#r for parameters)",
                        default="polymer_model_a#a_k#k_n#n_m#m_r#r.crs")
    parser.add_argument("-z", metavar='output zipped',
                        help="Should the output be a zipped directory. (True or False)",
                        choices=TRUTH_STRINGS,
                        default="False")
    parser.add_argument("-of", metavar="output format",
                        help="file format to be written. e.g. '.crs'",
                        choices=OUTPUT_FILE_FORMATS,
                        default=None)
    parser.add_argument("-rn", metavar="reaction notation",
                        help="Output reaction notation", default="FULL")
    parser.add_argument("-an", metavar="arrow notation",
                        help="Output arrow notation", default="USES_EQUALS")
    parser.add_argument("-ow", metavar="overwrite ok", help="Sets if the program is allowed to " +
                        "write over files", choices=TRUTH_STRINGS, default="False")

    arguments = vars(parser.parse_args())

    alphabet_sizes = Ut.parse_integer_input(arguments["a"])
    food_max_lengths = Ut.parse_integer_input(arguments["k"])
    polymer_max_lengths = Ut.parse_integer_input(arguments["n"])
    means = Ut.parse_float_input(arguments["m"])
    number_of_replicates = Ut.parse_integer_input(arguments["r"])
    zipped = True if arguments['z'].casefold(
    ) in ['True'.casefold(), "1"] else False
    overwrite_ok = True if arguments['ow'].casefold(
    ) in ['True'.casefold(), "1"] else False

    cpn.generate_reaction_system_files(alphabet_sizes,
                                       food_max_lengths,
                                       polymer_max_lengths,
                                       means,
                                       number_of_replicates,
                                       arguments['o'],
                                       arguments['f'],
                                       zipped,
                                       arguments['of'],
                                       arguments['rn'],
                                       arguments['an'],
                                       overwrite_ok)


if __name__ == "__main__":
    main()



from __future__ import annotations
from tqdm import tqdm
import random
import copy
from algorithm.MaxRAFAlgorithm import MaxRAFAlgorithm
from algorithm.MinIRAFHeuristic import MinIRAFHeuristic
from algorithm.CoreRAFAlgorithm import CoreRAFAlgorithm
from algorithm.AlgorithmBase import AlgorithmBase
from model.ReactionSystem import ReactionSystem, MoleculeType
from model.Reaction import Reaction, FORMAL_FOOD
from time import time

import sys
import os
sys.path.insert(0, os.path.abspath(
    os.path.join(os.path.dirname(__file__), '.')))


class MinRAFGeneratingElement(AlgorithmBase):

    def __init__(self, targets:list[MoleculeType] = [])-> MinRAFGeneratingElement:
        self.min_i_raf_heuristic = MinIRAFHeuristic()
        self.targets:set[MoleculeType] = targets
    
    NAME: str = "Min RAF Generating Element"
    
    @property
    def number_of_random_insertion_orders(self):
        return self.min_i_raf_heuristic.number_of_random_insertion_orders

    @number_of_random_insertion_orders.setter
    def number_of_random_insertion_orders(self, value: int):
        self.min_i_raf_heuristic.number_of_random_insertion_orders = value
    
    @property
    def name(self):
        return self.NAME

    @name.setter
    def name(self, value: str):
        self.NAME = value

    @property
    def description(self):
        return ("Identifies a subset of the maxRAF that is (i) a RAF and (ii) generates a " + 
                "given element x (not in the food set) and (iii) which is minimal amongst " +
                "all such sets satisfying (i) and (ii)")
        
    def apply(self, input:ReactionSystem) -> ReactionSystem:
        result_sys_name = ""
        if len(self.targets) == 1:
            result_sys_name += self.name + "'" + list(self.targets)[0].name + "'"
        else:
            result_sys_name += self.name + "'" + str(len(self.targets)) + "targets'"
            
        empty_sys = ReactionSystem(result_sys_name)
        
        if not self.targets:
            tqdm.write("No targets selected")
            return empty_sys
        elif input.foods.intersection(self.targets):
            tqdm.write("A target element is contained in the food set.")
            return empty_sys
        
        max_raf = MaxRAFAlgorithm().apply(input)
        if not max_raf.reactions:
            tqdm.write("Max RAF is empty.")
            return empty_sys
        
        augmented_sys = ReactionSystem()
        augmented_sys.foods = max_raf.foods.copy()
        for reaction in tqdm(max_raf.reactions, desc="Augmenting Reactions: "):
            augmented_reaction:Reaction = copy.deepcopy(reaction)
            for target in self.targets:
                augmented_reaction.catalysts = self.add_item_to_all(
                    augmented_reaction.catalyst_conjunctions, target)
            augmented_sys.reactions.append(augmented_reaction)
        
        i_raf = self.min_i_raf_heuristic.apply(augmented_sys)
        i_raf.name = result_sys_name
        if i_raf.size == 0:
            tqdm.write("Irreducible RAF is empty.")
            return empty_sys
        
        core_raf = CoreRAFAlgorithm().apply(augmented_sys)
        if core_raf.size > 0:
            tqdm.write("Irreducible is unique.")
            
        return i_raf
    
    def add_item_to_all(self, conjunctions:set[MoleculeType], item:MoleculeType):
        if not conjunctions or all([conj == FORMAL_FOOD for conj in conjunctions]):
            conjunctions.clear()
            return item.name
        else:
            res = []
            for conj in conjunctions:
                if conj == item: res.append(item)
                else: res.append(MoleculeType().value_of(conj.name + "&" + item.name))
            
            return ",".join(res)